import os
from unittest import mock

import boto3
import pytest
from botocore.exceptions import ClientError

from moto import mock_aws, settings


def create_vpx_pcx(ec2, client):
    vpc = ec2.create_vpc(CidrBlock="10.0.0.0/16")
    peer_vpc = ec2.create_vpc(CidrBlock="11.0.0.0/16")
    vpc_pcx = client.create_vpc_peering_connection(VpcId=vpc.id, PeerVpcId=peer_vpc.id)
    vpc_pcx = vpc_pcx["VpcPeeringConnection"]
    return vpc_pcx


@mock_aws
def test_vpc_peering_connections_boto3():
    ec2 = boto3.resource("ec2", region_name="us-east-1")
    client = boto3.client("ec2", region_name="us-east-1")

    vpc_pcx = create_vpx_pcx(ec2, client)

    assert "VpcPeeringConnectionId" in vpc_pcx
    assert vpc_pcx["Status"]["Code"] == "initiating-request"


@mock_aws
def test_vpc_peering_connections_get_all_boto3():
    ec2 = boto3.resource("ec2", region_name="us-east-1")
    client = boto3.client("ec2", region_name="us-east-1")
    vpc_pcx = create_vpx_pcx(ec2, client)
    vpc_pcx_id = vpc_pcx["VpcPeeringConnectionId"]

    all_vpc_pcxs = retrieve_all(client)
    assert vpc_pcx_id in [vpc_pcx["VpcPeeringConnectionId"] for vpc_pcx in all_vpc_pcxs]
    my_vpc_pcx = [
        vpc_pcx
        for vpc_pcx in all_vpc_pcxs
        if vpc_pcx["VpcPeeringConnectionId"] == vpc_pcx_id
    ][0]
    assert my_vpc_pcx["Status"]["Code"] == "pending-acceptance"
    assert my_vpc_pcx["Status"]["Message"] == "Pending Acceptance by 123456789012"


@mock_aws
def test_vpc_peering_connections_accept_boto3():
    ec2 = boto3.resource("ec2", region_name="us-east-1")
    client = boto3.client("ec2", region_name="us-east-1")
    vpc_pcx = create_vpx_pcx(ec2, client)
    vpc_pcx_id = vpc_pcx["VpcPeeringConnectionId"]

    vpc_pcx = client.accept_vpc_peering_connection(VpcPeeringConnectionId=vpc_pcx_id)
    vpc_pcx = vpc_pcx["VpcPeeringConnection"]
    assert vpc_pcx["Status"]["Code"] == "active"

    with pytest.raises(ClientError) as ex:
        client.reject_vpc_peering_connection(VpcPeeringConnectionId=vpc_pcx_id)
    assert ex.value.response["ResponseMetadata"]["HTTPStatusCode"] == 400
    assert "RequestId" in ex.value.response["ResponseMetadata"]
    assert ex.value.response["Error"]["Code"] == "InvalidStateTransition"

    my_vpc_pcxs = client.describe_vpc_peering_connections(
        VpcPeeringConnectionIds=[vpc_pcx_id]
    )["VpcPeeringConnections"]
    assert len(my_vpc_pcxs) == 1
    assert my_vpc_pcxs[0]["Status"]["Code"] == "active"


@mock_aws
def test_vpc_peering_connections_reject_boto3():
    ec2 = boto3.resource("ec2", region_name="us-east-1")
    client = boto3.client("ec2", region_name="us-east-1")
    vpc_pcx = create_vpx_pcx(ec2, client)
    vpc_pcx_id = vpc_pcx["VpcPeeringConnectionId"]

    client.reject_vpc_peering_connection(VpcPeeringConnectionId=vpc_pcx_id)

    with pytest.raises(ClientError) as ex:
        client.accept_vpc_peering_connection(VpcPeeringConnectionId=vpc_pcx_id)
    assert ex.value.response["ResponseMetadata"]["HTTPStatusCode"] == 400
    assert "RequestId" in ex.value.response["ResponseMetadata"]
    assert ex.value.response["Error"]["Code"] == "InvalidStateTransition"

    my_pcxs = client.describe_vpc_peering_connections(
        VpcPeeringConnectionIds=[vpc_pcx_id]
    )["VpcPeeringConnections"]
    assert len(my_pcxs) == 1
    assert my_pcxs[0]["Status"]["Code"] == "rejected"


@mock_aws
def test_vpc_peering_connections_delete_boto3():
    ec2 = boto3.resource("ec2", region_name="us-east-1")
    client = boto3.client("ec2", region_name="us-east-1")
    vpc_pcx = create_vpx_pcx(ec2, client)
    vpc_pcx_id = vpc_pcx["VpcPeeringConnectionId"]

    client.delete_vpc_peering_connection(VpcPeeringConnectionId=vpc_pcx_id)

    all_vpc_pcxs = retrieve_all(client)
    assert vpc_pcx_id in [vpcx["VpcPeeringConnectionId"] for vpcx in all_vpc_pcxs]

    my_vpcx = [
        vpcx for vpcx in all_vpc_pcxs if vpcx["VpcPeeringConnectionId"] == vpc_pcx_id
    ][0]
    assert my_vpcx["Status"]["Code"] == "deleted"

    with pytest.raises(ClientError) as ex:
        client.delete_vpc_peering_connection(VpcPeeringConnectionId="pcx-1234abcd")
    assert ex.value.response["ResponseMetadata"]["HTTPStatusCode"] == 400
    assert "RequestId" in ex.value.response["ResponseMetadata"]
    assert (
        ex.value.response["Error"]["Code"] == "InvalidVpcPeeringConnectionId.NotFound"
    )


@mock_aws
@pytest.mark.parametrize(
    "account1,account2",
    [
        pytest.param("111111111111", "111111111111", id="within account"),
        pytest.param("111111111111", "222222222222", id="across accounts"),
    ],
)
@pytest.mark.skipif(
    settings.TEST_SERVER_MODE, reason="Cannot set account ID in server mode"
)
def test_vpc_peering_connections_cross_region(account1, account2):
    # create vpc in us-west-1 and ap-northeast-1
    with mock.patch.dict(os.environ, {"MOTO_ACCOUNT_ID": account1}):
        ec2_usw1 = boto3.resource("ec2", region_name="us-west-1")
        vpc_usw1 = ec2_usw1.create_vpc(CidrBlock="10.90.0.0/16")

    with mock.patch.dict(os.environ, {"MOTO_ACCOUNT_ID": account2}):
        ec2_apn1 = boto3.resource("ec2", region_name="ap-northeast-1")
        vpc_apn1 = ec2_apn1.create_vpc(CidrBlock="10.20.0.0/16")

    # create peering
    with mock.patch.dict(os.environ, {"MOTO_ACCOUNT_ID": account1}):
        ec2_usw1 = boto3.client("ec2", region_name="us-west-1")
        vpc_pcx_usw1 = ec2_usw1.create_vpc_peering_connection(
            VpcId=vpc_usw1.id,
            PeerVpcId=vpc_apn1.id,
            PeerRegion="ap-northeast-1",
            PeerOwnerId=account2,
        )

    assert (
        vpc_pcx_usw1["VpcPeeringConnection"]["Status"]["Code"] == "initiating-request"
    )
    assert (
        vpc_pcx_usw1["VpcPeeringConnection"]["Status"]["Message"]
        == f"Initiating Request to {account2}"
    )
    assert (
        vpc_pcx_usw1["VpcPeeringConnection"]["RequesterVpcInfo"]["VpcId"] == vpc_usw1.id
    )
    assert (
        vpc_pcx_usw1["VpcPeeringConnection"]["RequesterVpcInfo"]["CidrBlock"]
        == "10.90.0.0/16"
    )
    assert (
        vpc_pcx_usw1["VpcPeeringConnection"]["RequesterVpcInfo"]["OwnerId"] == account1
    )
    assert (
        vpc_pcx_usw1["VpcPeeringConnection"]["RequesterVpcInfo"]["Region"]
        == "us-west-1"
    )
    assert (
        vpc_pcx_usw1["VpcPeeringConnection"]["AccepterVpcInfo"]["VpcId"] == vpc_apn1.id
    )
    assert (
        vpc_pcx_usw1["VpcPeeringConnection"]["AccepterVpcInfo"]["CidrBlock"]
        == "10.20.0.0/16"
    )
    assert (
        vpc_pcx_usw1["VpcPeeringConnection"]["AccepterVpcInfo"]["OwnerId"] == account2
    )
    assert (
        vpc_pcx_usw1["VpcPeeringConnection"]["AccepterVpcInfo"]["Region"]
        == "ap-northeast-1"
    )

    # test cross region vpc peering connection exist
    with mock.patch.dict(os.environ, {"MOTO_ACCOUNT_ID": account2}):
        vpc_pcx_apn1 = ec2_apn1.VpcPeeringConnection(
            vpc_pcx_usw1["VpcPeeringConnection"]["VpcPeeringConnectionId"]
        )
        assert (
            vpc_pcx_apn1.id
            == vpc_pcx_usw1["VpcPeeringConnection"]["VpcPeeringConnectionId"]
        )
        assert vpc_pcx_apn1.requester_vpc.id == vpc_usw1.id
        assert vpc_pcx_apn1.accepter_vpc.id == vpc_apn1.id

    # Quick check to verify the options have a default value
    accepter_options = vpc_pcx_apn1.accepter_vpc_info["PeeringOptions"]
    assert accepter_options["AllowDnsResolutionFromRemoteVpc"] is False
    assert accepter_options["AllowEgressFromLocalClassicLinkToRemoteVpc"] is False
    assert accepter_options["AllowEgressFromLocalVpcToRemoteClassicLink"] is False
    requester_options = vpc_pcx_apn1.requester_vpc_info["PeeringOptions"]
    assert requester_options["AllowDnsResolutionFromRemoteVpc"] is False
    assert requester_options["AllowEgressFromLocalClassicLinkToRemoteVpc"] is False
    assert requester_options["AllowEgressFromLocalVpcToRemoteClassicLink"] is False


@mock_aws
@pytest.mark.parametrize(
    "account1,account2",
    [
        pytest.param("111111111111", "111111111111", id="within account"),
        pytest.param("111111111111", "222222222222", id="across accounts"),
    ],
)
@pytest.mark.skipif(
    settings.TEST_SERVER_MODE, reason="Cannot set account ID in server mode"
)
def test_modify_vpc_peering_connections_accepter_only(account1, account2):
    # create vpc in us-west-1 and ap-northeast-1
    with mock.patch.dict(os.environ, {"MOTO_ACCOUNT_ID": account1}):
        ec2_usw1 = boto3.resource("ec2", region_name="us-west-1")
        client = boto3.client("ec2", region_name="us-west-1")
        vpc_usw1 = ec2_usw1.create_vpc(CidrBlock="10.90.0.0/16")

    with mock.patch.dict(os.environ, {"MOTO_ACCOUNT_ID": account2}):
        ec2_apn1 = boto3.resource("ec2", region_name="ap-northeast-1")
        vpc_apn1 = ec2_apn1.create_vpc(CidrBlock="10.20.0.0/16")

    # create peering
    with mock.patch.dict(os.environ, {"MOTO_ACCOUNT_ID": account1}):
        vpc_pcx_usw1 = ec2_usw1.create_vpc_peering_connection(
            VpcId=vpc_usw1.id,
            PeerVpcId=vpc_apn1.id,
            PeerRegion="ap-northeast-1",
            PeerOwnerId=account2,
        )

        # modify peering connection options
        client.modify_vpc_peering_connection_options(
            VpcPeeringConnectionId=vpc_pcx_usw1.id,
            AccepterPeeringConnectionOptions={"AllowDnsResolutionFromRemoteVpc": True},
        )

        # Accepter options are different
        vpc_pcx_usw1.reload()

    accepter_options = vpc_pcx_usw1.accepter_vpc_info["PeeringOptions"]
    assert accepter_options["AllowDnsResolutionFromRemoteVpc"] is True
    assert accepter_options["AllowEgressFromLocalClassicLinkToRemoteVpc"] is False
    assert accepter_options["AllowEgressFromLocalVpcToRemoteClassicLink"] is False

    # Requester options are untouched
    requester_options = vpc_pcx_usw1.requester_vpc_info["PeeringOptions"]
    assert requester_options["AllowDnsResolutionFromRemoteVpc"] is False
    assert requester_options["AllowEgressFromLocalClassicLinkToRemoteVpc"] is False
    assert requester_options["AllowEgressFromLocalVpcToRemoteClassicLink"] is False


@mock_aws
@pytest.mark.parametrize(
    "account1,account2",
    [
        pytest.param("111111111111", "111111111111", id="within account"),
        pytest.param("111111111111", "222222222222", id="across accounts"),
    ],
)
@pytest.mark.skipif(
    settings.TEST_SERVER_MODE, reason="Cannot set account ID in server mode"
)
def test_modify_vpc_peering_connections_requester_only(account1, account2):
    # create vpc in us-west-1 and ap-northeast-1
    with mock.patch.dict(os.environ, {"MOTO_ACCOUNT_ID": account1}):
        ec2_usw1 = boto3.resource("ec2", region_name="us-west-1")
        client = boto3.client("ec2", region_name="us-west-1")
        vpc_usw1 = ec2_usw1.create_vpc(CidrBlock="10.90.0.0/16")

    with mock.patch.dict(os.environ, {"MOTO_ACCOUNT_ID": account2}):
        ec2_apn1 = boto3.resource("ec2", region_name="ap-northeast-1")
        vpc_apn1 = ec2_apn1.create_vpc(CidrBlock="10.20.0.0/16")

    # create peering
    with mock.patch.dict(os.environ, {"MOTO_ACCOUNT_ID": account1}):
        vpc_pcx_usw1 = ec2_usw1.create_vpc_peering_connection(
            VpcId=vpc_usw1.id,
            PeerVpcId=vpc_apn1.id,
            PeerRegion="ap-northeast-1",
            PeerOwnerId=account2,
        )
        #
        client.modify_vpc_peering_connection_options(
            VpcPeeringConnectionId=vpc_pcx_usw1.id,
            RequesterPeeringConnectionOptions={
                "AllowEgressFromLocalVpcToRemoteClassicLink": True,
            },
        )
        # Requester options are different
        vpc_pcx_usw1.reload()

    requester_options = vpc_pcx_usw1.requester_vpc_info["PeeringOptions"]
    assert requester_options["AllowDnsResolutionFromRemoteVpc"] is False
    assert requester_options["AllowEgressFromLocalClassicLinkToRemoteVpc"] is False
    assert requester_options["AllowEgressFromLocalVpcToRemoteClassicLink"] is True

    # Accepter options are untouched
    accepter_options = vpc_pcx_usw1.accepter_vpc_info["PeeringOptions"]
    assert accepter_options["AllowDnsResolutionFromRemoteVpc"] is False
    assert accepter_options["AllowEgressFromLocalClassicLinkToRemoteVpc"] is False
    assert accepter_options["AllowEgressFromLocalVpcToRemoteClassicLink"] is False


@mock_aws
@pytest.mark.parametrize(
    "account1,account2",
    [
        pytest.param("111111111111", "111111111111", id="within account"),
        pytest.param("111111111111", "222222222222", id="across accounts"),
    ],
)
@pytest.mark.skipif(
    settings.TEST_SERVER_MODE, reason="Cannot set account ID in server mode"
)
def test_modify_vpc_peering_connections_unknown_vpc(account1, account2):
    # create vpc in us-west-1 and ap-northeast-1
    with mock.patch.dict(os.environ, {"MOTO_ACCOUNT_ID": account1}):
        ec2_usw1 = boto3.resource("ec2", region_name="us-west-1")
        client = boto3.client("ec2", region_name="us-west-1")
        vpc_usw1 = ec2_usw1.create_vpc(CidrBlock="10.90.0.0/16")

    with mock.patch.dict(os.environ, {"MOTO_ACCOUNT_ID": account2}):
        ec2_apn1 = boto3.resource("ec2", region_name="ap-northeast-1")
        vpc_apn1 = ec2_apn1.create_vpc(CidrBlock="10.20.0.0/16")

    # create peering
    with mock.patch.dict(os.environ, {"MOTO_ACCOUNT_ID": account1}):
        ec2_usw1.create_vpc_peering_connection(
            VpcId=vpc_usw1.id,
            PeerVpcId=vpc_apn1.id,
            PeerRegion="ap-northeast-1",
            PeerOwnerId=account2,
        )

        with pytest.raises(ClientError) as ex:
            client.modify_vpc_peering_connection_options(
                VpcPeeringConnectionId="vpx-unknown",
                RequesterPeeringConnectionOptions={},
            )
    err = ex.value.response["Error"]
    assert err["Code"] == "InvalidVpcPeeringConnectionId.NotFound"
    assert err["Message"] == "VpcPeeringConnectionID vpx-unknown does not exist."


@mock_aws
@pytest.mark.parametrize(
    "account1,account2",
    [
        pytest.param("111111111111", "111111111111", id="within account"),
        pytest.param("111111111111", "222222222222", id="across accounts"),
    ],
)
@pytest.mark.skipif(
    settings.TEST_SERVER_MODE, reason="Cannot set account ID in server mode"
)
def test_vpc_peering_connections_cross_region_fail(account1, account2):
    # create vpc in us-west-1 and ap-northeast-1
    with mock.patch.dict(os.environ, {"MOTO_ACCOUNT_ID": account1}):
        ec2_usw1 = boto3.resource("ec2", region_name="us-west-1")
        vpc_usw1 = ec2_usw1.create_vpc(CidrBlock="10.90.0.0/16")

    with mock.patch.dict(os.environ, {"MOTO_ACCOUNT_ID": account2}):
        ec2_apn1 = boto3.resource("ec2", region_name="ap-northeast-1")
        vpc_apn1 = ec2_apn1.create_vpc(CidrBlock="10.20.0.0/16")

    # create peering wrong region with no vpc
    with pytest.raises(ClientError) as cm:
        with mock.patch.dict(os.environ, {"MOTO_ACCOUNT_ID": account1}):
            ec2_usw1.create_vpc_peering_connection(
                VpcId=vpc_usw1.id,
                PeerVpcId=vpc_apn1.id,
                PeerRegion="ap-northeast-2",
                PeerOwnerId=account2,
            )
    assert cm.value.response["Error"]["Code"] == "InvalidVpcID.NotFound"


@mock_aws
@pytest.mark.parametrize(
    "account1,account2",
    [
        pytest.param("111111111111", "111111111111", id="within account"),
        pytest.param("111111111111", "222222222222", id="across accounts"),
    ],
)
@pytest.mark.skipif(
    settings.TEST_SERVER_MODE, reason="Cannot set account ID in server mode"
)
def test_describe_vpc_peering_connections_only_returns_requested_id(account1, account2):
    # create vpc in us-west-1 and ap-northeast-1
    with mock.patch.dict(os.environ, {"MOTO_ACCOUNT_ID": account1}):
        ec2_usw1 = boto3.resource("ec2", region_name="us-west-1")
        vpc_usw1 = ec2_usw1.create_vpc(CidrBlock="10.90.0.0/16")

    with mock.patch.dict(os.environ, {"MOTO_ACCOUNT_ID": account2}):
        ec2_apn1 = boto3.resource("ec2", region_name="ap-northeast-1")
        vpc_apn1 = ec2_apn1.create_vpc(CidrBlock="10.20.0.0/16")

    # create peering
    with mock.patch.dict(os.environ, {"MOTO_ACCOUNT_ID": account1}):
        vpc_pcx_usw1 = ec2_usw1.create_vpc_peering_connection(
            VpcId=vpc_usw1.id,
            PeerVpcId=vpc_apn1.id,
            PeerRegion="ap-northeast-1",
            PeerOwnerId=account2,
        )
        vpc_pcx_usw2 = ec2_usw1.create_vpc_peering_connection(
            VpcId=vpc_usw1.id,
            PeerVpcId=vpc_apn1.id,
            PeerRegion="ap-northeast-1",
            PeerOwnerId=account2,
        )

        # describe peering
        ec2_usw1 = boto3.client("ec2", region_name="us-west-1")
        our_vpcx = [vpcx["VpcPeeringConnectionId"] for vpcx in retrieve_all(ec2_usw1)]

        assert vpc_pcx_usw1.id in our_vpcx
        assert vpc_pcx_usw2.id in our_vpcx
        assert vpc_apn1.id not in our_vpcx

        both_pcx = ec2_usw1.describe_vpc_peering_connections(
            VpcPeeringConnectionIds=[vpc_pcx_usw1.id, vpc_pcx_usw2.id]
        )["VpcPeeringConnections"]
        assert len(both_pcx) == 2

        one_pcx = ec2_usw1.describe_vpc_peering_connections(
            VpcPeeringConnectionIds=[vpc_pcx_usw1.id]
        )["VpcPeeringConnections"]
        assert len(one_pcx) == 1


@mock_aws
@pytest.mark.parametrize(
    "account1,account2",
    [
        pytest.param("111111111111", "111111111111", id="within account"),
        pytest.param("111111111111", "222222222222", id="across accounts"),
    ],
)
@pytest.mark.skipif(
    settings.TEST_SERVER_MODE, reason="Cannot set account ID in server mode"
)
def test_vpc_peering_connections_cross_region_accept(account1, account2):
    # create vpc in us-west-1 and ap-northeast-1
    with mock.patch.dict(os.environ, {"MOTO_ACCOUNT_ID": account1}):
        ec2_usw1 = boto3.resource("ec2", region_name="us-west-1")
        vpc_usw1 = ec2_usw1.create_vpc(CidrBlock="10.90.0.0/16")

    with mock.patch.dict(os.environ, {"MOTO_ACCOUNT_ID": account2}):
        ec2_apn1 = boto3.resource("ec2", region_name="ap-northeast-1")
        vpc_apn1 = ec2_apn1.create_vpc(CidrBlock="10.20.0.0/16")

    # create peering
    with mock.patch.dict(os.environ, {"MOTO_ACCOUNT_ID": account1}):
        vpc_pcx_usw1 = ec2_usw1.create_vpc_peering_connection(
            VpcId=vpc_usw1.id,
            PeerVpcId=vpc_apn1.id,
            PeerRegion="ap-northeast-1",
            PeerOwnerId=account2,
        )

    # accept peering from ap-northeast-1
    with mock.patch.dict(os.environ, {"MOTO_ACCOUNT_ID": account2}):
        ec2_apn1 = boto3.client("ec2", region_name="ap-northeast-1")
        acp_pcx_apn1 = ec2_apn1.accept_vpc_peering_connection(
            VpcPeeringConnectionId=vpc_pcx_usw1.id
        )

    with mock.patch.dict(os.environ, {"MOTO_ACCOUNT_ID": account1}):
        ec2_usw1 = boto3.client("ec2", region_name="us-west-1")
        des_pcx_apn1 = ec2_usw1.describe_vpc_peering_connections(
            VpcPeeringConnectionIds=[vpc_pcx_usw1.id]
        )
        des_pcx_usw1 = ec2_usw1.describe_vpc_peering_connections(
            VpcPeeringConnectionIds=[vpc_pcx_usw1.id]
        )

    assert acp_pcx_apn1["VpcPeeringConnection"]["Status"]["Code"] == "active"
    assert (
        acp_pcx_apn1["VpcPeeringConnection"]["AccepterVpcInfo"]["Region"]
        == "ap-northeast-1"
    )
    assert (
        acp_pcx_apn1["VpcPeeringConnection"]["RequesterVpcInfo"]["Region"]
        == "us-west-1"
    )
    assert des_pcx_apn1["VpcPeeringConnections"][0]["Status"]["Code"] == "active"
    assert (
        des_pcx_apn1["VpcPeeringConnections"][0]["AccepterVpcInfo"]["Region"]
        == "ap-northeast-1"
    )
    assert (
        des_pcx_apn1["VpcPeeringConnections"][0]["RequesterVpcInfo"]["Region"]
        == "us-west-1"
    )
    assert des_pcx_usw1["VpcPeeringConnections"][0]["Status"]["Code"] == "active"
    assert (
        des_pcx_usw1["VpcPeeringConnections"][0]["AccepterVpcInfo"]["Region"]
        == "ap-northeast-1"
    )
    assert (
        des_pcx_usw1["VpcPeeringConnections"][0]["RequesterVpcInfo"]["Region"]
        == "us-west-1"
    )


@mock_aws
@pytest.mark.parametrize(
    "account1,account2",
    [
        pytest.param("111111111111", "111111111111", id="within account"),
        pytest.param("111111111111", "222222222222", id="across accounts"),
    ],
)
@pytest.mark.skipif(
    settings.TEST_SERVER_MODE, reason="Cannot set account ID in server mode"
)
def test_vpc_peering_connections_cross_region_reject(account1, account2):
    # create vpc in us-west-1 and ap-northeast-1
    with mock.patch.dict(os.environ, {"MOTO_ACCOUNT_ID": account1}):
        ec2_usw1 = boto3.resource("ec2", region_name="us-west-1")
        vpc_usw1 = ec2_usw1.create_vpc(CidrBlock="10.90.0.0/16")

    with mock.patch.dict(os.environ, {"MOTO_ACCOUNT_ID": account2}):
        ec2_apn1 = boto3.resource("ec2", region_name="ap-northeast-1")
        vpc_apn1 = ec2_apn1.create_vpc(CidrBlock="10.20.0.0/16")

    # create peering
    with mock.patch.dict(os.environ, {"MOTO_ACCOUNT_ID": account1}):
        vpc_pcx_usw1 = ec2_usw1.create_vpc_peering_connection(
            VpcId=vpc_usw1.id,
            PeerVpcId=vpc_apn1.id,
            PeerRegion="ap-northeast-1",
            PeerOwnerId=account2,
        )

    # reject peering from ap-northeast-1
    with mock.patch.dict(os.environ, {"MOTO_ACCOUNT_ID": account2}):
        ec2_apn1 = boto3.client("ec2", region_name="ap-northeast-1")
        rej_pcx_apn1 = ec2_apn1.reject_vpc_peering_connection(
            VpcPeeringConnectionId=vpc_pcx_usw1.id
        )

    with mock.patch.dict(os.environ, {"MOTO_ACCOUNT_ID": account1}):
        ec2_usw1 = boto3.client("ec2", region_name="us-west-1")
        des_pcx_apn1 = ec2_usw1.describe_vpc_peering_connections(
            VpcPeeringConnectionIds=[vpc_pcx_usw1.id]
        )
        des_pcx_usw1 = ec2_usw1.describe_vpc_peering_connections(
            VpcPeeringConnectionIds=[vpc_pcx_usw1.id]
        )
    assert rej_pcx_apn1["Return"] is True
    assert des_pcx_apn1["VpcPeeringConnections"][0]["Status"]["Code"] == "rejected"
    assert des_pcx_usw1["VpcPeeringConnections"][0]["Status"]["Code"] == "rejected"


@mock_aws
@pytest.mark.parametrize(
    "account1,account2",
    [
        pytest.param("111111111111", "111111111111", id="within account"),
        pytest.param("111111111111", "222222222222", id="across accounts"),
    ],
)
@pytest.mark.skipif(
    settings.TEST_SERVER_MODE, reason="Cannot set account ID in server mode"
)
def test_vpc_peering_connections_cross_region_delete(account1, account2):
    # create vpc in us-west-1 and ap-northeast-1
    with mock.patch.dict(os.environ, {"MOTO_ACCOUNT_ID": account1}):
        ec2_usw1 = boto3.resource("ec2", region_name="us-west-1")
        vpc_usw1 = ec2_usw1.create_vpc(CidrBlock="10.90.0.0/16")

    with mock.patch.dict(os.environ, {"MOTO_ACCOUNT_ID": account2}):
        ec2_apn1 = boto3.resource("ec2", region_name="ap-northeast-1")
        vpc_apn1 = ec2_apn1.create_vpc(CidrBlock="10.20.0.0/16")

    # create peering
    with mock.patch.dict(os.environ, {"MOTO_ACCOUNT_ID": account1}):
        vpc_pcx_usw1 = ec2_usw1.create_vpc_peering_connection(
            VpcId=vpc_usw1.id,
            PeerVpcId=vpc_apn1.id,
            PeerRegion="ap-northeast-1",
            PeerOwnerId=account2,
        )

    with mock.patch.dict(os.environ, {"MOTO_ACCOUNT_ID": account2}):
        # reject peering from ap-northeast-1
        ec2_apn1 = boto3.client("ec2", region_name="ap-northeast-1")
        del_pcx_apn1 = ec2_apn1.delete_vpc_peering_connection(
            VpcPeeringConnectionId=vpc_pcx_usw1.id
        )

    with mock.patch.dict(os.environ, {"MOTO_ACCOUNT_ID": account1}):
        ec2_usw1 = boto3.client("ec2", region_name="us-west-1")
        des_pcx_apn1 = ec2_usw1.describe_vpc_peering_connections(
            VpcPeeringConnectionIds=[vpc_pcx_usw1.id]
        )
        des_pcx_usw1 = ec2_usw1.describe_vpc_peering_connections(
            VpcPeeringConnectionIds=[vpc_pcx_usw1.id]
        )

    assert del_pcx_apn1["Return"] is True
    assert des_pcx_apn1["VpcPeeringConnections"][0]["Status"]["Code"] == "deleted"
    assert (
        des_pcx_apn1["VpcPeeringConnections"][0]["Status"]["Message"]
        == f"Deleted by {account2}"
    )

    assert des_pcx_usw1["VpcPeeringConnections"][0]["Status"]["Code"] == "deleted"
    assert (
        des_pcx_usw1["VpcPeeringConnections"][0]["Status"]["Message"]
        == f"Deleted by {account2}"
    )


@mock_aws
@pytest.mark.parametrize(
    "account1,account2",
    [
        pytest.param("111111111111", "111111111111", id="within account"),
        pytest.param("111111111111", "222222222222", id="across accounts"),
    ],
)
@pytest.mark.skipif(
    settings.TEST_SERVER_MODE, reason="Cannot set account ID in server mode"
)
def test_vpc_peering_connections_cross_region_accept_wrong_region(account1, account2):
    # create vpc in us-west-1 and ap-northeast-1
    with mock.patch.dict(os.environ, {"MOTO_ACCOUNT_ID": account1}):
        ec2_usw1 = boto3.resource("ec2", region_name="us-west-1")
        vpc_usw1 = ec2_usw1.create_vpc(CidrBlock="10.90.0.0/16")

    with mock.patch.dict(os.environ, {"MOTO_ACCOUNT_ID": account2}):
        ec2_apn1 = boto3.resource("ec2", region_name="ap-northeast-1")
        vpc_apn1 = ec2_apn1.create_vpc(CidrBlock="10.20.0.0/16")

    # create peering
    with mock.patch.dict(os.environ, {"MOTO_ACCOUNT_ID": account1}):
        vpc_pcx_usw1 = ec2_usw1.create_vpc_peering_connection(
            VpcId=vpc_usw1.id,
            PeerVpcId=vpc_apn1.id,
            PeerRegion="ap-northeast-1",
            PeerOwnerId=account2,
        )

    # accept wrong peering from us-west-1 which will raise error
    # only applicable for cross-region intra-account peering.
    if account1 == account2:
        with mock.patch.dict(os.environ, {"MOTO_ACCOUNT_ID": account1}):
            ec2_usw1 = boto3.client("ec2", region_name="us-west-1")
            with pytest.raises(ClientError) as cm:
                ec2_usw1.accept_vpc_peering_connection(
                    VpcPeeringConnectionId=vpc_pcx_usw1.id
                )

        assert cm.value.response["Error"]["Code"] == "OperationNotPermitted"
        exp_msg = f"Incorrect region (us-west-1) specified for this request. VPC peering connection {vpc_pcx_usw1.id} must be accepted in region ap-northeast-1"
        assert cm.value.response["Error"]["Message"] == exp_msg

    # Ensure accepting peering from requester account raises
    # only applicable for cross-region inter-account peering.
    if account1 != account2:
        with mock.patch.dict(os.environ, {"MOTO_ACCOUNT_ID": account1}):
            ec2_usw1 = boto3.client("ec2", region_name="us-west-1")
            with pytest.raises(ClientError) as cm:
                ec2_usw1.accept_vpc_peering_connection(
                    VpcPeeringConnectionId=vpc_pcx_usw1.id
                )

        assert cm.value.response["Error"]["Code"] == "OperationNotPermitted"
        exp_msg = f"User ({account1}) cannot accept peering {vpc_pcx_usw1.id}"
        assert cm.value.response["Error"]["Message"] == exp_msg


@mock_aws
@pytest.mark.parametrize(
    "account1,account2",
    [
        pytest.param("111111111111", "111111111111", id="within account"),
        pytest.param("111111111111", "222222222222", id="across accounts"),
    ],
)
@pytest.mark.skipif(
    settings.TEST_SERVER_MODE, reason="Cannot set account ID in server mode"
)
def test_vpc_peering_connections_cross_region_reject_wrong_region(account1, account2):
    # create vpc in us-west-1 and ap-northeast-1
    with mock.patch.dict(os.environ, {"MOTO_ACCOUNT_ID": account1}):
        ec2_usw1 = boto3.resource("ec2", region_name="us-west-1")
        vpc_usw1 = ec2_usw1.create_vpc(CidrBlock="10.90.0.0/16")

    with mock.patch.dict(os.environ, {"MOTO_ACCOUNT_ID": account2}):
        ec2_apn1 = boto3.resource("ec2", region_name="ap-northeast-1")
        vpc_apn1 = ec2_apn1.create_vpc(CidrBlock="10.20.0.0/16")

    # create peering
    with mock.patch.dict(os.environ, {"MOTO_ACCOUNT_ID": account1}):
        vpc_pcx_usw1 = ec2_usw1.create_vpc_peering_connection(
            VpcId=vpc_usw1.id,
            PeerVpcId=vpc_apn1.id,
            PeerRegion="ap-northeast-1",
            PeerOwnerId=account2,
        )

    # reject wrong peering from us-west-1 which will raise error.
    # only applicable for cross-region intra-account peering.
    if account1 == account2:
        with mock.patch.dict(os.environ, {"MOTO_ACCOUNT_ID": account1}):
            ec2_usw1 = boto3.client("ec2", region_name="us-west-1")
            with pytest.raises(ClientError) as cm:
                ec2_usw1.reject_vpc_peering_connection(
                    VpcPeeringConnectionId=vpc_pcx_usw1.id
                )

        assert cm.value.response["Error"]["Code"] == "OperationNotPermitted"
        exp_msg = f"Incorrect region (us-west-1) specified for this request. VPC peering connection {vpc_pcx_usw1.id} must be accepted or rejected in region ap-northeast-1"
        assert cm.value.response["Error"]["Message"] == exp_msg

    # Ensure rejecting peering from requester account raises
    # only applicable for cross-region inter-account peering.
    if account1 != account2:
        with mock.patch.dict(os.environ, {"MOTO_ACCOUNT_ID": account1}):
            ec2_usw1 = boto3.client("ec2", region_name="us-west-1")
            with pytest.raises(ClientError) as cm:
                ec2_usw1.reject_vpc_peering_connection(
                    VpcPeeringConnectionId=vpc_pcx_usw1.id
                )

        assert cm.value.response["Error"]["Code"] == "OperationNotPermitted"
        exp_msg = f"User ({account1}) cannot reject peering {vpc_pcx_usw1.id}"
        assert cm.value.response["Error"]["Message"] == exp_msg


def retrieve_all(client):
    resp = client.describe_vpc_peering_connections()
    all_vpx = resp["VpcPeeringConnections"]
    token = resp.get("NextToken")
    while token:
        resp = client.describe_vpc_peering_connections(NextToken=token)
        all_vpx.extend(resp["VpcPeeringConnections"])
        token = resp.get("NextToken")
    return all_vpx



from .responses import LambdaResponse

url_bases = [r"https?://lambda\.(.+)\.amazonaws\.com"]


url_paths = {
    r"{0}/(?P<api_version>[^/]+)/functions$": LambdaResponse.dispatch,
    r"{0}/(?P<api_version>[^/]+)/functions/$": LambdaResponse.dispatch,
    r"{0}/(?P<api_version>[^/]+)/functions/(?P<function_name>[\w_:%-]+)/?$": LambdaResponse.dispatch,
    r"{0}/(?P<api_version>[^/]+)/functions/(?P<function_name>[\w_:%-]+)/aliases$": LambdaResponse.dispatch,
    r"{0}/(?P<api_version>[^/]+)/functions/(?P<function_name>[\w_:%-]+)/aliases/(?P<alias_name>[\w_-]+)$": LambdaResponse.dispatch,
    r"{0}/(?P<api_version>[^/]+)/functions/(?P<function_name>[\w_:%-]+)/versions/?$": LambdaResponse.dispatch,
    r"{0}/(?P<api_version>[^/]+)/event-source-mappings/$": LambdaResponse.dispatch,
    r"{0}/(?P<api_version>[^/]+)/event-source-mappings/(?P<UUID>[\w_-]+)/?$": LambdaResponse.dispatch,
    r"{0}/(?P<api_version>[^/]+)/functions/(?P<function_name>[\w_-]+)/invocations/?$": LambdaResponse.dispatch,
    r"{0}/(?P<api_version>[^/]+)/functions/(?P<resource_arn>.+)/invocations/?$": LambdaResponse.dispatch,
    r"{0}/(?P<api_version>[^/]+)/functions/(?P<function_name>[\w_:%-]+)/invoke-async$": LambdaResponse.dispatch,
    r"{0}/(?P<api_version>[^/]+)/functions/(?P<function_name>[\w_:%-]+)/invoke-async/$": LambdaResponse.dispatch,
    r"{0}/(?P<api_version>[^/]+)/tags/(?P<resource_arn>.+)": LambdaResponse.dispatch,
    r"{0}/(?P<api_version>[^/]+)/functions/(?P<function_name>[\w_:%-]+)/policy/(?P<statement_id>[\w_-]+)$": LambdaResponse.dispatch,
    r"{0}/(?P<api_version>[^/]+)/functions/(?P<function_name>[\w_:%-]+)/policy/?$": LambdaResponse.dispatch,
    r"{0}/(?P<api_version>[^/]+)/functions/(?P<function_name>[\w_:%-]+)/configuration/?$": LambdaResponse.dispatch,
    r"{0}/(?P<api_version>[^/]+)/functions/(?P<function_name>[\w_:%-]+)/code/?$": LambdaResponse.dispatch,
    r"{0}/(?P<api_version>[^/]+)/functions/(?P<function_name>[\w_:%-]+)/code-signing-config$": LambdaResponse.dispatch,
    r"{0}/(?P<api_version>[^/]+)/functions/(?P<function_name>[\w_:%-]+)/concurrency/?$": LambdaResponse.dispatch,
    r"{0}/(?P<api_version>[^/]+)/functions/(?P<function_name>[\w_:%-]+)/url/?$": LambdaResponse.dispatch,
    r"{0}/(?P<api_version>[^/]+)/layers$": LambdaResponse.dispatch,
    r"{0}/(?P<api_version>[^/]+)/layers/$": LambdaResponse.dispatch,
    r"{0}/(?P<api_version>[^/]+)/layers/(?P<layer_name>.+)/versions$": LambdaResponse.dispatch,
    r"{0}/(?P<api_version>[^/]+)/layers/(?P<layer_name>.+)/versions/$": LambdaResponse.dispatch,
    r"{0}/(?P<api_version>[^/]+)/layers/(?P<layer_name>.+)/versions/(?P<layer_version>[\w_-]+)$": LambdaResponse.dispatch,
    r"{0}/(?P<api_version>[^/]+)/functions/(?P<function_name>[\w_:%-]+)/event-invoke-config/?$": LambdaResponse.dispatch,
    r"{0}/(?P<api_version>[^/]+)/functions/(?P<function_name>[\w_:%-]+)/event-invoke-config/list$": LambdaResponse.dispatch,
}



import boto3
import pytest
from botocore.exceptions import ClientError

from moto import mock_aws
from moto.core import DEFAULT_ACCOUNT_ID

# See our Development Tips on writing tests for hints on how to write good tests:
# http://docs.getmoto.org/en/latest/docs/contributing/development_tips/tests.html


@mock_aws
def test_create_get_delete_schedule_group():
    client = boto3.client("scheduler", region_name="eu-west-1")
    arn = client.create_schedule_group(Name="sg")["ScheduleGroupArn"]

    assert arn == f"arn:aws:scheduler:eu-west-1:{DEFAULT_ACCOUNT_ID}:schedule-group/sg"

    group = client.get_schedule_group(Name="sg")
    assert group["Arn"] == arn
    assert group["Name"] == "sg"
    assert group["State"] == "ACTIVE"

    client.delete_schedule_group(Name="sg")

    with pytest.raises(ClientError) as exc:
        client.get_schedule_group(Name="sg")
    err = exc.value.response["Error"]
    assert err["Code"] == "ResourceNotFoundException"


@mock_aws
def test_list_schedule_groups():
    client = boto3.client("scheduler", region_name="ap-southeast-1")

    # The default group is always active
    groups = client.list_schedule_groups()["ScheduleGroups"]
    assert len(groups) == 1
    assert (
        groups[0]["Arn"]
        == f"arn:aws:scheduler:ap-southeast-1:{DEFAULT_ACCOUNT_ID}:schedule-group/default"
    )

    arn1 = client.create_schedule_group(Name="sg")["ScheduleGroupArn"]

    groups = client.list_schedule_groups()["ScheduleGroups"]
    assert len(groups) == 2
    assert groups[1]["Arn"] == arn1



"""Exceptions raised by the timestreamquery service."""

from moto.core.exceptions import JsonRESTError


class ResourceNotFound(JsonRESTError):
    def __init__(self, arn: str):
        super().__init__(
            error_type="ResourceNotFoundException",
            message=f"The resource with arn {arn} does not exist.",
        )



import boto3
import pytest
from botocore.exceptions import ClientError

from moto import mock_aws

region = "eu-west-1"


@mock_aws
def test_put_object():
    client = boto3.client("mediastore-data", region_name=region)
    response = client.put_object(Body="011001", Path="foo")
    assert response["ResponseMetadata"]["HTTPStatusCode"] == 200
    assert response["StorageClass"] == "TEMPORAL"

    items = client.list_items()["Items"]
    assert any(d["Name"] == "foo" for d in items)


@mock_aws
def test_get_object_throws_not_found_error():
    client = boto3.client("mediastore-data", region_name=region)
    with pytest.raises(ClientError) as ex:
        client.get_object(Path="foo")
    assert ex.value.response["Error"]["Code"] == "ObjectNotFoundException"


@mock_aws
def test_get_object():
    client = boto3.client("mediastore-data", region_name=region)
    client.put_object(Body="011001", Path="foo")
    response = client.get_object(Path="foo")
    assert response["ResponseMetadata"]["HTTPStatusCode"] == 200
    assert response["ResponseMetadata"]["HTTPHeaders"]["path"] == "foo"
    assert response["Body"].read() == b"011001"


@mock_aws
def test_delete_object_error():
    client = boto3.client("mediastore-data", region_name=region)
    with pytest.raises(ClientError) as ex:
        client.delete_object(Path="foo")
    assert ex.value.response["Error"]["Code"] == "ObjectNotFoundException"


@mock_aws
def test_delete_object_succeeds():
    client = boto3.client("mediastore-data", region_name=region)
    object_path = "foo"
    client.put_object(Body="011001", Path=object_path)
    response = client.delete_object(Path=object_path)
    assert response["ResponseMetadata"]["HTTPStatusCode"] == 200

    assert len(client.list_items()["Items"]) == 0


@mock_aws
def test_list_items():
    client = boto3.client("mediastore-data", region_name=region)
    assert len(client.list_items()["Items"]) == 0

    object_path = "foo"
    client.put_object(Body="011001", Path=object_path)

    items = client.list_items()["Items"]
    assert len(items) == 1
    assert any(d["Name"] == object_path for d in items)



"""Test the different server responses."""

import json

import moto.server as server


def test_list_global_networks():
    backend = server.create_backend_app("networkmanager")
    test_client = backend.test_client()

    res = test_client.get("/global-networks")

    assert "GlobalNetworks" in json.loads(res.data)


def test_list_core_networks():
    backend = server.create_backend_app("networkmanager")
    test_client = backend.test_client()

    res = test_client.get("/core-networks")

    assert "CoreNetworks" in json.loads(res.data)


def test_tag_resource():
    backend = server.create_backend_app("networkmanager")
    test_client = backend.test_client()

    res = test_client.post(
        "/tags/test-resource-id",
        json={"Tags": [{"Key": "Name", "Value": "CoreNetworks"}]},
    )
    data = json.loads(res.data)
    assert data["Message"] == "Resource not found."
    assert data["ResourceId"] == "test-resource-id"



import datetime
import json
from typing import Any
from urllib.parse import urlparse

from moto.core.common_types import TYPE_RESPONSE
from moto.core.responses import BaseResponse
from moto.core.utils import utcnow


class InstanceMetadataResponse(BaseResponse):
    def __init__(self) -> None:
        super().__init__(service_name=None)

    def backends(self) -> None:
        pass

    @staticmethod
    def metadata_response(  # type: ignore
        request: Any,  # pylint: disable=unused-argument
        full_url: str,
        headers: Any,
    ) -> TYPE_RESPONSE:
        """
        Mock response for localhost metadata

        http://docs.aws.amazon.com/AWSEC2/latest/UserGuide/AESDG-chapter-instancedata.html
        """

        parsed_url = urlparse(full_url)
        tomorrow = utcnow() + datetime.timedelta(days=1)
        credentials = dict(
            AccessKeyId="test-key",
            SecretAccessKey="test-secret-key",
            Token="test-session-token",
            Expiration=tomorrow.strftime("%Y-%m-%dT%H:%M:%SZ"),
        )

        path = parsed_url.path

        meta_data_prefix = "/latest/meta-data/"
        # Strip prefix if it is there
        if path.startswith(meta_data_prefix):
            path = path[len(meta_data_prefix) :]

        if path == "":
            result = "iam"
        elif path == "iam":
            result = json.dumps({"security-credentials": {"default-role": credentials}})
        elif path == "iam/security-credentials/":
            result = "default-role"
        elif path == "iam/security-credentials/default-role":
            result = json.dumps(credentials)
        else:
            raise NotImplementedError(
                f"The {path} metadata path has not been implemented"
            )
        return 200, headers, result



from datetime import datetime, timedelta, timezone
from typing import Dict, List

import boto3
import pytest
from botocore.exceptions import ClientError

from moto import mock_aws


@pytest.fixture(name="domain_parameters")
def generate_domain_parameters() -> Dict:
    return {
        "DomainName": "domain.com",
        "DurationInYears": 3,
        "AutoRenew": True,
        "AdminContact": {
            "FirstName": "First",
            "LastName": "Last",
            "ContactType": "PERSON",
            "AddressLine1": "address 1",
            "AddressLine2": "address 2",
            "City": "New York City",
            "CountryCode": "US",
            "ZipCode": "123123123",
            "Email": "email@gmail.com",
            "Fax": "+1.1234567890",
        },
        "RegistrantContact": {
            "FirstName": "First",
            "LastName": "Last",
            "ContactType": "PERSON",
            "AddressLine1": "address 1",
            "AddressLine2": "address 2",
            "City": "New York City",
            "CountryCode": "US",
            "ZipCode": "123123123",
            "Email": "email@gmail.com",
            "Fax": "+1.1234567890",
        },
        "TechContact": {
            "FirstName": "First",
            "LastName": "Last",
            "ContactType": "PERSON",
            "AddressLine1": "address 1",
            "AddressLine2": "address 2",
            "City": "New York City",
            "CountryCode": "US",
            "ZipCode": "123123123",
            "Email": "email@gmail.com",
            "Fax": "+1.1234567890",
        },
        "PrivacyProtectAdminContact": True,
        "PrivacyProtectRegistrantContact": True,
        "PrivacyProtectTechContact": True,
    }


@pytest.fixture(name="invalid_domain_parameters")
def generate_invalid_domain_parameters(domain_parameters: Dict) -> Dict:
    domain_parameters["DomainName"] = "a"
    domain_parameters["DurationInYears"] = 500
    return domain_parameters


@mock_aws
def test_register_domain(domain_parameters: Dict):
    route53domains_client = boto3.client("route53domains", region_name="global")
    res = route53domains_client.register_domain(**domain_parameters)

    operation_id = res["OperationId"]

    operations = route53domains_client.list_operations(Type=["REGISTER_DOMAIN"])[
        "Operations"
    ]
    for operation in operations:
        if operation["OperationId"] == operation_id:
            return

    assert (
        operation_id in [operation["OperationId"] for operation in operations]
    ), "Could not find expected operation id returned from `register_domain` in operation list"


@mock_aws
def test_register_domain_creates_hosted_zone(
    domain_parameters: Dict,
):
    """Test good register domain API calls."""
    route53domains_client = boto3.client("route53domains", region_name="global")
    route53_client = boto3.client("route53", region_name="global")
    route53domains_client.register_domain(**domain_parameters)

    res = route53_client.list_hosted_zones()
    assert "domain.com" in [
        zone["Name"] for zone in res["HostedZones"]
    ], "`register_domain` did not create a new hosted zone with the same name"


@mock_aws
def test_register_domain_fails_on_invalid_input(
    invalid_domain_parameters: Dict,
):
    route53domains_client = boto3.client("route53domains", region_name="global")
    route53_client = boto3.client("route53", region_name="global")
    with pytest.raises(ClientError) as exc:
        route53domains_client.register_domain(**invalid_domain_parameters)

    err = exc.value.response["Error"]
    assert err["Code"] == "InvalidInput"

    res = route53_client.list_hosted_zones()
    assert len(res["HostedZones"]) == 0


@mock_aws
def test_register_domain_fails_on_invalid_tld(domain_parameters: Dict):
    route53domains_client = boto3.client("route53domains", region_name="global")
    route53_client = boto3.client("route53", region_name="global")

    params = domain_parameters.copy()
    params["DomainName"] = "test.non-existing-tld"
    with pytest.raises(ClientError) as exc:
        route53domains_client.register_domain(**params)

    err = exc.value.response["Error"]
    assert err["Code"] == "UnsupportedTLD"

    res = route53_client.list_hosted_zones()
    assert len(res["HostedZones"]) == 0


@mock_aws
def test_list_operations(domain_parameters: Dict):
    route53domains_client = boto3.client("route53domains", region_name="global")
    route53domains_client.register_domain(**domain_parameters)
    operations = route53domains_client.list_operations()["Operations"]
    assert len(operations) == 1

    future_time = datetime.now(timezone.utc) + timedelta(minutes=1)
    operations = route53domains_client.list_operations(
        SubmittedSince=future_time.timestamp()
    )["Operations"]
    assert len(operations) == 0

    operations = route53domains_client.list_operations(Status=["SUCCESSFUL"])[
        "Operations"
    ]
    assert len(operations) == 1

    operations = route53domains_client.list_operations(Status=["IN_PROGRESS"])[
        "Operations"
    ]
    assert len(operations) == 0

    operations = route53domains_client.list_operations(Type=["REGISTER_DOMAIN"])[
        "Operations"
    ]
    assert len(operations) == 1

    operations = route53domains_client.list_operations(Type=["DELETE_DOMAIN"])[
        "Operations"
    ]
    assert len(operations) == 0


@mock_aws
def test_list_operations_invalid_input():
    route53domains_client = boto3.client("route53domains", region_name="global")
    with pytest.raises(ClientError) as exc:
        _ = route53domains_client.list_operations(Type=["INVALID_TYPE"])["Operations"]

    err = exc.value.response["Error"]
    assert err["Code"] == "InvalidInput"

    with pytest.raises(ClientError) as exc:
        _ = route53domains_client.list_operations(Status=["INVALID_STATUS"])[
            "Operations"
        ]
    err = exc.value.response["Error"]
    assert err["Code"] == "InvalidInput"


@mock_aws
def test_get_operation_detail(domain_parameters: Dict):
    route53domains_client = boto3.client("route53domains", region_name="global")
    res = route53domains_client.register_domain(**domain_parameters)
    expected_operation_id = res["OperationId"]
    operation = route53domains_client.get_operation_detail(
        OperationId=expected_operation_id
    )
    assert operation["OperationId"] == expected_operation_id
    assert operation["Status"] == "SUCCESSFUL"
    assert operation["Type"] == "REGISTER_DOMAIN"


@mock_aws
def test_get_nonexistent_operation_detail():
    route53domains_client = boto3.client("route53domains", region_name="global")
    with pytest.raises(ClientError) as exc:
        route53domains_client.get_operation_detail(
            OperationId="non-exiting-operation-id"
        )

    err = exc.value.response["Error"]
    assert err["Code"] == "InvalidInput"


@mock_aws
def test_duplicate_requests(domain_parameters: Dict):
    route53domains_client = boto3.client("route53domains", region_name="global")
    route53domains_client.register_domain(**domain_parameters)
    with pytest.raises(ClientError) as exc:
        route53domains_client.register_domain(**domain_parameters)
    err = exc.value.response["Error"]
    assert err["Code"] == "DuplicateRequest"


@mock_aws
def test_domain_limit(domain_parameters: Dict):
    route53domains_client = boto3.client("route53domains", region_name="global")
    params = domain_parameters.copy()
    for i in range(20):
        params["DomainName"] = f"domain-{i}.com"
        route53domains_client.register_domain(**params)

    params["DomainName"] = "domain-20.com"
    with pytest.raises(ClientError) as exc:
        route53domains_client.register_domain(**params)

    err = exc.value.response["Error"]
    assert err["Code"] == "DomainLimitExceeded"


@mock_aws
def test_get_domain_detail(domain_parameters: Dict):
    route53domains_client = boto3.client("route53domains", region_name="global")
    route53domains_client.register_domain(**domain_parameters)
    res = route53domains_client.get_domain_detail(
        DomainName=domain_parameters["DomainName"]
    )
    assert res["DomainName"] == domain_parameters["DomainName"]


@mock_aws
def test_get_invalid_domain_detail(domain_parameters):
    route53domains_client = boto3.client("route53domains", region_name="global")
    route53domains_client.register_domain(**domain_parameters)

    with pytest.raises(ClientError) as exc:
        route53domains_client.get_domain_detail(DomainName="not-a-domain")

    err = exc.value.response["Error"]
    assert err["Code"] == "InvalidInput"

    with pytest.raises(ClientError) as exc:
        route53domains_client.get_domain_detail(DomainName="test.non-existing-tld")

    err = exc.value.response["Error"]
    assert err["Code"] == "UnsupportedTLD"


@mock_aws
def test_list_domains(domain_parameters: Dict):
    route53domains_client = boto3.client("route53domains", region_name="global")
    route53domains_client.register_domain(**domain_parameters)
    res = route53domains_client.list_domains()

    assert len(res["Domains"]) == 1
    params = domain_parameters.copy()
    params["DomainName"] = "new-domain.com"
    route53domains_client.register_domain(**params)
    res = route53domains_client.list_domains()
    assert len(res["Domains"]) == 2


@mock_aws
@pytest.mark.parametrize(
    "filters,expected_domains_len",
    [
        (
            [
                {
                    "Name": "DomainName",
                    "Operator": "BEGINS_WITH",
                    "Values": ["some-non-registered-domain.com"],
                }
            ],
            0,  # expected_domains_len
        ),
        (
            [
                {
                    "Name": "DomainName",
                    "Operator": "BEGINS_WITH",
                    "Values": ["domain.com"],
                }
            ],
            1,  # expected_domains_len
        ),
        (
            [
                {
                    "Name": "Expiry",
                    "Operator": "GE",
                    "Values": [
                        str(
                            datetime.fromisocalendar(
                                year=2012, week=20, day=3
                            ).timestamp()
                        )
                    ],
                }
            ],
            1,  # expected_domains_len
        ),
        (
            [
                {
                    "Name": "Expiry",
                    "Operator": "GE",
                    "Values": [
                        str(
                            datetime.fromisocalendar(
                                year=2050, week=20, day=3
                            ).timestamp()
                        )
                    ],
                }
            ],
            0,  # expected_domains_len
        ),
    ],
)
def test_list_domains_filters(
    domain_parameters: Dict, filters: List[Dict], expected_domains_len: int
):
    route53domains_client = boto3.client("route53domains", region_name="global")
    route53domains_client.register_domain(**domain_parameters)
    res = route53domains_client.list_domains(FilterConditions=filters)
    assert len(res["Domains"]) == expected_domains_len


@mock_aws
def test_list_domains_sort_condition(domain_parameters: Dict):
    route53domains_client = boto3.client("route53domains", region_name="global")
    params = domain_parameters.copy()
    params["DomainName"] = "adomain.com"
    route53domains_client.register_domain(**params)
    params["DomainName"] = "bdomain.com"
    route53domains_client.register_domain(**params)
    sort = {"Name": "DomainName", "SortOrder": "DES"}
    res = route53domains_client.list_domains(SortCondition=sort)
    domains = res["Domains"]
    assert domains[0]["DomainName"] == "bdomain.com"
    assert domains[1]["DomainName"] == "adomain.com"

    sort = {"Name": "Expiry", "SortOrder": "ASC"}
    res = route53domains_client.list_domains(SortCondition=sort)
    domains = res["Domains"]
    assert domains[0]["DomainName"] == "adomain.com"
    assert domains[1]["DomainName"] == "bdomain.com"


@mock_aws
def test_list_domains_invalid_filter(domain_parameters: Dict):
    route53domains_client = boto3.client("route53domains", region_name="global")
    route53domains_client.register_domain(**domain_parameters)
    filters = [
        {
            "Name": "InvalidField",
            "Operator": "InvalidOperator",
            "Values": ["value-1", "value-2"],  # multiple values isn't supported
        }
    ]

    with pytest.raises(ClientError) as exc:
        route53domains_client.list_domains(FilterConditions=filters)

    err = exc.value.response["Error"]
    assert err["Code"] == "InvalidInput"


@mock_aws
def test_list_domains_invalid_sort_condition(domain_parameters: Dict):
    route53domains_client = boto3.client("route53domains", region_name="global")
    route53domains_client.register_domain(**domain_parameters)
    sort = {
        "Name": "InvalidField",
        "SortOrder": "InvalidOrder",
    }

    with pytest.raises(ClientError) as exc:
        route53domains_client.list_domains(SortCondition=sort)

    err = exc.value.response["Error"]
    assert err["Code"] == "InvalidInput"


@mock_aws
def test_list_domains_sort_condition_not_the_same_as_filter_condition(
    domain_parameters: Dict,
):
    route53domains_client = boto3.client("route53domains", region_name="global")
    route53domains_client.register_domain(**domain_parameters)
    sort = {
        "Name": "Expiry",
        "SortOrder": "ASC",
    }
    filters = [
        {
            "Name": "DomainName",
            "Operator": "BEGINS_WITH",
            "Values": ["domain.com"],
        }
    ]

    with pytest.raises(ClientError) as exc:
        route53domains_client.list_domains(FilterConditions=filters, SortCondition=sort)

    err = exc.value.response["Error"]
    assert err["Code"] == "InvalidInput"


@mock_aws
def test_delete_domain(domain_parameters: Dict):
    route53domains_client = boto3.client("route53domains", region_name="global")
    route53domains_client.register_domain(**domain_parameters)
    domains = route53domains_client.list_domains()["Domains"]
    assert len(domains) == 1
    route53domains_client.delete_domain(DomainName=domain_parameters["DomainName"])
    domains = route53domains_client.list_domains()["Domains"]
    assert len(domains) == 0
    operations = route53domains_client.list_operations(Type=["DELETE_DOMAIN"])[
        "Operations"
    ]
    assert len(operations) == 1


@mock_aws
def test_delete_invalid_domain(domain_parameters: Dict):
    route53domains_client = boto3.client("route53domains", region_name="global")
    domains = route53domains_client.list_domains()["Domains"]
    assert len(domains) == 0
    with pytest.raises(ClientError) as exc:
        route53domains_client.delete_domain(DomainName=domain_parameters["DomainName"])

    err = exc.value.response["Error"]
    assert err["Code"] == "InvalidInput"


@mock_aws
@pytest.mark.parametrize(
    "nameservers",
    [
        [{"Name": "1-nameserver.net"}, {"Name": "2-nameserver.net"}],
        [
            {"Name": "3-nameserver.net", "GlueIps": ["1.1.1.2"]},
            {"Name": "4-nameserver.net", "GlueIps": ["1.1.1.1"]},
        ],
    ],
)
def test_update_domain_nameservers(domain_parameters: Dict, nameservers: List[Dict]):
    route53domains_client = boto3.client("route53domains", region_name="global")
    route53domains_client.register_domain(**domain_parameters)
    operation_id = route53domains_client.update_domain_nameservers(
        DomainName=domain_parameters["DomainName"], Nameservers=nameservers
    )["OperationId"]
    domain = route53domains_client.get_domain_detail(
        DomainName=domain_parameters["DomainName"]
    )
    assert domain["Nameservers"] == nameservers
    operation = route53domains_client.get_operation_detail(OperationId=operation_id)
    assert operation["Type"] == "UPDATE_NAMESERVER"
    assert operation["Status"] == "SUCCESSFUL"


@mock_aws
@pytest.mark.parametrize(
    "nameservers",
    [
        [{"Name": "1-nameserver.net", "GlueIps": ["1.1.1.1", "1.1.1.2"]}],
        [
            {
                "Name": "1-nameserver.net",
                "GlueIps": [
                    "2001:0db8:85a3:0000:0000:8a2e:0370:7334",
                    "2001:0db8:85a3:0000:0000:8a2e:0370:7334",
                ],
            }
        ],
        [
            {
                "Name": "1-nameserver.net",
                "GlueIps": [
                    "2001:0db8:85a3:0000:0000:8a2e:0370:7334",
                    "2001:0db8:85a3:0000:0000:8a2e:0370:7334",
                    "1.1.1.1",
                ],
            }
        ],
        [
            {
                "Name": "1-nameserver.net",
                "GlueIps": [
                    "2001:0db8:85a3:0000:0000:8a2e:0370:7334",
                    "2001:0db8:85a3:0000:0000:8a2e:0370:7334",
                    "1.1.1.1",
                    "1.1.1.2",
                ],
            }
        ],
        [
            {
                "Name": "1-nameserver.net",
                "GlueIps": [
                    "2001:0db8:85a3:0000:0000:8a2e:0370:7334",
                    "1.1.1.1",
                    "not-an-ip-address",
                ],
            }
        ],
    ],
)
def test_update_domain_nameservers_with_multiple_glue_ips(
    domain_parameters: Dict, nameservers: List[Dict]
):
    route53domains_client = boto3.client("route53domains", region_name="global")
    route53domains_client.register_domain(**domain_parameters)
    with pytest.raises(ClientError) as exc:
        route53domains_client.update_domain_nameservers(
            DomainName=domain_parameters["DomainName"], Nameservers=nameservers
        )
    err = exc.value.response["Error"]
    assert err["Code"] == "InvalidInput"


@mock_aws
def test_update_domain_nameservers_requires_glue_ips(domain_parameters: Dict):
    route53domains_client = boto3.client("route53domains", region_name="global")
    route53domains_client.register_domain(**domain_parameters)
    domain_name = domain_parameters["DomainName"]
    nameservers = [{"Name": f"subdomain.{domain_name}"}]
    with pytest.raises(ClientError) as exc:
        route53domains_client.update_domain_nameservers(
            DomainName=domain_parameters["DomainName"], Nameservers=nameservers
        )
    err = exc.value.response["Error"]
    assert err["Code"] == "InvalidInput"


@mock_aws
def test_update_domain_nameservers_for_nonexistent_domain():
    route53domains_client = boto3.client("route53domains", region_name="global")
    nameservers = [{"Name": "1-nameserver.net"}, {"Name": "2-nameserver.net"}]

    with pytest.raises(ClientError) as exc:
        route53domains_client.update_domain_nameservers(
            DomainName="non-existent-domain.com", Nameservers=nameservers
        )
    err = exc.value.response["Error"]
    assert err["Code"] == "InvalidInput"



"""Unit tests for glue-supported APIs."""

from random import randint
from uuid import uuid4

import boto3
import pytest
from botocore.client import ClientError
from botocore.exceptions import ParamValidationError

from moto import mock_aws


@mock_aws
def test_create_job():
    client = create_glue_client()
    job_name = str(uuid4())
    response = client.create_job(
        Name=job_name, Role="test_role", Command=dict(Name="test_command")
    )
    assert response["Name"] == job_name


@mock_aws
def test_delete_job():
    client = create_glue_client()
    job_name = create_test_job(client)

    client.get_job(JobName=job_name)

    client.delete_job(JobName=job_name)

    with pytest.raises(ClientError) as exc:
        client.get_job(JobName=job_name)

    assert exc.value.response["Error"]["Code"] == "EntityNotFoundException"


@mock_aws
def test_create_job_default_argument_not_provided():
    client = create_glue_client()
    with pytest.raises(ParamValidationError) as exc:
        client.create_job(Role="test_role", Command=dict(Name="test_command"))

    assert exc.value.kwargs["report"] == 'Missing required parameter in input: "Name"'


@mock_aws
def test_list_jobs():
    client = create_glue_client()
    expected_jobs = randint(1, 15)
    create_test_jobs(client, expected_jobs)
    response = client.list_jobs()
    assert len(response["JobNames"]) == expected_jobs
    assert "NextToken" not in response


@mock_aws
def test_get_job_not_exists():
    client = create_glue_client()
    name = "my_job_name"

    with pytest.raises(ClientError) as exc:
        client.get_job(JobName=name)

    assert exc.value.response["Error"]["Code"] == "EntityNotFoundException"
    assert exc.value.response["Error"]["Message"] == "Job my_job_name not found."


@mock_aws
def test_get_job_exists():
    client = create_glue_client()
    job_attributes = {
        "Description": "test_description",
        "LogUri": "test_log/",
        "Role": "test_role",
        "ExecutionProperty": {"MaxConcurrentRuns": 123},
        "Command": {
            "Name": "test_command",
            "ScriptLocation": "test_s3_path",
            "PythonVersion": "3.6",
        },
        "DefaultArguments": {"string": "string"},
        "NonOverridableArguments": {"string": "string"},
        "Connections": {
            "Connections": [
                "string",
            ]
        },
        "MaxRetries": 123,
        "AllocatedCapacity": 123,
        "Timeout": 123,
        "MaxCapacity": 123.0,
        "WorkerType": "G.2X",
        "NumberOfWorkers": 123,
        "SecurityConfiguration": "test_config",
        "NotificationProperty": {"NotifyDelayAfter": 123},
        "GlueVersion": "string",
        "CodeGenConfigurationNodes": {},
        "ExecutionClass": "string",
        "SourceControlDetails": {},
    }
    job_name = create_test_job_w_all_attributes(client, **job_attributes)
    job = client.get_job(JobName=job_name)["Job"]
    assert job["Name"] == job_name
    assert "Description" in job
    assert "LogUri" in job
    assert "Role" in job
    assert job["ExecutionProperty"] == {"MaxConcurrentRuns": 123}
    assert "CreatedOn" in job
    assert "LastModifiedOn" in job
    assert "ExecutionProperty" in job
    assert "Command" in job
    assert "DefaultArguments" in job
    assert "NonOverridableArguments" in job
    assert "Connections" in job
    assert "MaxRetries" in job
    assert "AllocatedCapacity" in job
    assert "Timeout" in job
    assert "MaxCapacity" in job
    assert "WorkerType" in job
    assert "NumberOfWorkers" in job
    assert "SecurityConfiguration" in job
    assert "NotificationProperty" in job
    assert "GlueVersion" in job
    assert "CodeGenConfigurationNodes" in job
    assert "ExecutionClass" in job
    assert "SourceControlDetails" in job


@mock_aws
def test_get_jobs_job_name_exists():
    client = create_glue_client()
    test_job_name = create_test_job(client)
    response = client.get_jobs()
    assert len(response["Jobs"]) == 1
    assert response["Jobs"][0]["Name"] == test_job_name


@mock_aws
def test_get_jobs_with_max_results():
    client = create_glue_client()
    create_test_jobs(client, 4)
    response = client.get_jobs(MaxResults=2)
    assert len(response["Jobs"]) == 2
    assert "NextToken" in response


@mock_aws
def test_get_jobs_from_next_token():
    client = create_glue_client()
    create_test_jobs(client, 10)
    first_response = client.get_jobs(MaxResults=3)
    response = client.get_jobs(NextToken=first_response["NextToken"])
    assert len(response["Jobs"]) == 7


@mock_aws
def test_get_jobs_with_max_results_greater_than_actual_results():
    client = create_glue_client()
    create_test_jobs(client, 4)
    response = client.get_jobs(MaxResults=10)
    assert len(response["Jobs"]) == 4


@mock_aws
def test_get_jobs_next_token_logic_does_not_create_infinite_loop():
    client = create_glue_client()
    create_test_jobs(client, 4)
    first_response = client.get_jobs(MaxResults=1)
    next_token = first_response["NextToken"]
    while next_token:
        response = client.get_jobs(NextToken=next_token)
        next_token = response.get("NextToken")
    assert not next_token


@mock_aws
def test_list_jobs_with_max_results():
    client = create_glue_client()
    create_test_jobs(client, 4)
    response = client.list_jobs(MaxResults=2)
    assert len(response["JobNames"]) == 2
    assert "NextToken" in response


@mock_aws
def test_list_jobs_from_next_token():
    client = create_glue_client()
    create_test_jobs(client, 10)
    first_response = client.list_jobs(MaxResults=3)
    response = client.list_jobs(NextToken=first_response["NextToken"])
    assert len(response["JobNames"]) == 7


@mock_aws
def test_list_jobs_with_max_results_greater_than_actual_results():
    client = create_glue_client()
    create_test_jobs(client, 4)
    response = client.list_jobs(MaxResults=10)
    assert len(response["JobNames"]) == 4


@mock_aws
def test_list_jobs_with_tags():
    client = create_glue_client()
    create_test_job(client)
    create_test_job(client, {"string": "string"})
    response = client.list_jobs(Tags={"string": "string"})
    assert len(response["JobNames"]) == 1


@mock_aws
def test_list_jobs_after_tagging():
    client = create_glue_client()
    job_name = create_test_job(client)
    resource_arn = f"arn:aws:glue:us-east-1:123456789012:job/{job_name}"

    client.tag_resource(ResourceArn=resource_arn, TagsToAdd={"key1": "value1"})

    response = client.list_jobs(Tags={"key1": "value1"})
    assert len(response["JobNames"]) == 1


@mock_aws
def test_list_jobs_after_removing_tag():
    client = create_glue_client()
    job_name = create_test_job(client, {"key1": "value1"})
    resource_arn = f"arn:aws:glue:us-east-1:123456789012:job/{job_name}"

    client.untag_resource(ResourceArn=resource_arn, TagsToRemove=["key1"])

    response = client.list_jobs(Tags={"key1": "value1"})
    assert len(response["JobNames"]) == 0


@mock_aws
def test_list_jobs_next_token_logic_does_not_create_infinite_loop():
    client = create_glue_client()
    create_test_jobs(client, 4)
    first_response = client.list_jobs(MaxResults=1)
    next_token = first_response["NextToken"]
    while next_token:
        response = client.list_jobs(NextToken=next_token)
        next_token = response.get("NextToken")
    assert not next_token


@mock_aws
def test_batch_get_jobs():
    client = create_glue_client()
    job_name = create_test_job(client)

    response = client.batch_get_jobs(JobNames=[job_name, "job-not-found"])

    assert len(response["Jobs"]) == 1
    assert len(response["JobsNotFound"]) == 1


def create_glue_client():
    return boto3.client("glue", region_name="us-east-1")


def create_test_job(client, tags=None):
    job_name = str(uuid4())
    client.create_job(
        Name=job_name,
        Role="test_role",
        Command=dict(Name="test_command"),
        Tags=tags or {},
    )
    return job_name


def create_test_job_w_all_attributes(client, **job_attributes):
    job_name = str(uuid4())
    client.create_job(Name=job_name, **job_attributes)
    return job_name


def create_test_jobs(client, number_of_jobs):
    for _ in range(number_of_jobs):
        create_test_job(client)


def create_test_crawler(client, tags=None):
    crawler_name = str(uuid4())
    client.create_crawler(
        Name=crawler_name,
        Role="test_role",
        Targets={"S3Targets": [{"Path": "s3://tests3target"}]},
        Tags=tags or {},
    )
    return crawler_name


def create_test_crawlers(client, number_of_crawlers):
    for _ in range(number_of_crawlers):
        create_test_crawler(client)


@mock_aws
def test_list_crawlers_with_max_results():
    client = create_glue_client()
    create_test_crawlers(client, 4)
    response = client.list_crawlers(MaxResults=2)
    assert len(response["CrawlerNames"]) == 2
    assert "NextToken" in response


@mock_aws
def test_list_crawlers_from_next_token():
    client = create_glue_client()
    create_test_crawlers(client, 10)
    first_response = client.list_crawlers(MaxResults=3)
    response = client.list_crawlers(NextToken=first_response["NextToken"])
    assert len(response["CrawlerNames"]) == 7


@mock_aws
def test_list_crawlers_with_max_results_greater_than_actual_results():
    client = create_glue_client()
    create_test_crawlers(client, 4)
    response = client.list_crawlers(MaxResults=10)
    assert len(response["CrawlerNames"]) == 4


@mock_aws
def test_list_crawlers_with_tags():
    client = create_glue_client()
    create_test_crawler(client)
    create_test_crawler(client, {"string": "string"})
    response = client.list_crawlers(Tags={"string": "string"})
    assert len(response["CrawlerNames"]) == 1


@mock_aws
def test_list_crawlers_after_tagging():
    client = create_glue_client()
    crawler_name = create_test_crawler(client)
    resource_arn = f"arn:aws:glue:us-east-1:123456789012:crawler/{crawler_name}"

    client.tag_resource(ResourceArn=resource_arn, TagsToAdd={"key1": "value1"})

    response = client.list_crawlers(Tags={"key1": "value1"})
    assert len(response["CrawlerNames"]) == 1


@mock_aws
def test_list_crawlers_after_removing_tag():
    client = create_glue_client()
    crawler_name = create_test_crawler(client, {"key1": "value1"})
    resource_arn = f"arn:aws:glue:us-east-1:123456789012:crawler/{crawler_name}"

    client.untag_resource(ResourceArn=resource_arn, TagsToRemove=["key1"])

    response = client.list_crawlers(Tags={"key1": "value1"})
    assert len(response["CrawlerNames"]) == 0


@mock_aws
def test_list_crawlers_next_token_logic_does_not_create_infinite_loop():
    client = create_glue_client()
    create_test_crawlers(client, 4)
    first_response = client.list_crawlers(MaxResults=1)
    next_token = first_response["NextToken"]
    while next_token:
        response = client.list_crawlers(NextToken=next_token)
        next_token = response.get("NextToken")
    assert not next_token


@mock_aws
def test_get_tags_job():
    client = create_glue_client()
    job_name = create_test_job(client, {"key1": "value1", "key2": "value2"})
    resource_arn = f"arn:aws:glue:us-east-1:123456789012:job/{job_name}"

    resp = client.get_tags(ResourceArn=resource_arn)

    assert resp["Tags"] == {"key1": "value1", "key2": "value2"}


@mock_aws
def test_get_tags_jobs_no_tags():
    client = create_glue_client()
    job_name = create_test_job(client)
    resource_arn = f"arn:aws:glue:us-east-1:123456789012:job/{job_name}"

    resp = client.get_tags(ResourceArn=resource_arn)

    assert resp["Tags"] == {}


@mock_aws
def test_tag_glue_job():
    client = create_glue_client()
    job_name = create_test_job(client)
    resource_arn = f"arn:aws:glue:us-east-1:123456789012:job/{job_name}"

    client.tag_resource(
        ResourceArn=resource_arn, TagsToAdd={"key1": "value1", "key2": "value2"}
    )

    resp = client.get_tags(ResourceArn=resource_arn)

    assert resp["Tags"] == {"key1": "value1", "key2": "value2"}


@mock_aws
def test_untag_glue_job():
    client = create_glue_client()
    job_name = create_test_job(client)
    resource_arn = f"arn:aws:glue:us-east-1:123456789012:job/{job_name}"

    client.tag_resource(
        ResourceArn=resource_arn,
        TagsToAdd={"key1": "value1", "key2": "value2", "key3": "value3"},
    )

    client.untag_resource(ResourceArn=resource_arn, TagsToRemove=["key2"])

    resp = client.get_tags(ResourceArn=resource_arn)

    assert resp["Tags"] == {"key1": "value1", "key3": "value3"}


@mock_aws
def test_get_tags_crawler():
    client = create_glue_client()
    crawler_name = create_test_crawler(client, {"key1": "value1", "key2": "value2"})
    resource_arn = f"arn:aws:glue:us-east-1:123456789012:crawler/{crawler_name}"

    resp = client.get_tags(ResourceArn=resource_arn)

    assert resp["Tags"] == {"key1": "value1", "key2": "value2"}


@mock_aws
def test_get_tags_crawler_no_tags():
    client = create_glue_client()
    crawler_name = create_test_crawler(client)
    resource_arn = f"arn:aws:glue:us-east-1:123456789012:crawler/{crawler_name}"

    resp = client.get_tags(ResourceArn=resource_arn)

    assert resp["Tags"] == {}


@mock_aws
def test_tag_glue_crawler():
    client = create_glue_client()
    crawler_name = create_test_crawler(client)
    resource_arn = f"arn:aws:glue:us-east-1:123456789012:crawler/{crawler_name}"

    client.tag_resource(
        ResourceArn=resource_arn, TagsToAdd={"key1": "value1", "key2": "value2"}
    )

    resp = client.get_tags(ResourceArn=resource_arn)

    assert resp["Tags"] == {"key1": "value1", "key2": "value2"}


@mock_aws
def test_untag_glue_crawler():
    client = create_glue_client()
    crawler_name = create_test_crawler(client)
    resource_arn = f"arn:aws:glue:us-east-1:123456789012:crawler/{crawler_name}"

    client.tag_resource(
        ResourceArn=resource_arn,
        TagsToAdd={"key1": "value1", "key2": "value2", "key3": "value3"},
    )

    client.untag_resource(ResourceArn=resource_arn, TagsToRemove=["key2"])

    resp = client.get_tags(ResourceArn=resource_arn)

    assert resp["Tags"] == {"key1": "value1", "key3": "value3"}


@mock_aws
def test_batch_get_crawlers():
    client = create_glue_client()
    crawler_name = create_test_crawler(client)

    response = client.batch_get_crawlers(
        CrawlerNames=[crawler_name, "crawler-not-found"]
    )

    assert len(response["Crawlers"]) == 1
    assert len(response["CrawlersNotFound"]) == 1


@mock_aws
def test_create_trigger():
    client = create_glue_client()
    job_name = create_test_job(client)
    trigger_name = str(uuid4())

    response = client.create_trigger(
        Name=trigger_name,
        Type="ON_DEMAND",
        Actions=[
            {
                "JobName": job_name,
            }
        ],
    )
    assert response["Name"] == trigger_name


@mock_aws
def test_get_trigger_on_demand():
    client = create_glue_client()
    job_name = create_test_job(client)
    trigger_name = str(uuid4())
    trigger_attributes = {
        "Type": "ON_DEMAND",
        "Actions": [
            {
                "JobName": job_name,
            }
        ],
    }
    client.create_trigger(Name=trigger_name, **trigger_attributes)

    trigger = client.get_trigger(Name=trigger_name)["Trigger"]

    assert trigger["Name"] == trigger_name
    assert trigger["Type"] == "ON_DEMAND"
    assert trigger["State"] == "CREATED"
    assert trigger["Actions"] == [{"JobName": job_name}]


@mock_aws
def test_get_trigger_scheduled():
    client = create_glue_client()
    job_name = create_test_job(client)
    trigger_name = str(uuid4())
    trigger_attributes = {
        "Type": "SCHEDULED",
        "Schedule": "cron(5 3 * * ? *)",
        "Actions": [
            {
                "JobName": job_name,
            }
        ],
        "StartOnCreation": True,
    }
    client.create_trigger(Name=trigger_name, **trigger_attributes)

    trigger = client.get_trigger(Name=trigger_name)["Trigger"]

    assert trigger["Name"] == trigger_name
    assert trigger["Type"] == "SCHEDULED"
    assert trigger["State"] == "ACTIVATED"
    assert trigger["Actions"] == [{"JobName": job_name}]


@mock_aws
def test_get_trigger_conditional():
    client = create_glue_client()
    crawler_name = create_test_crawler(client)
    job_name = create_test_job(client)
    trigger_name = str(uuid4())
    trigger_attributes = {
        "Type": "CONDITIONAL",
        "Actions": [
            {
                "JobName": job_name,
            }
        ],
        "StartOnCreation": True,
        "Predicate": {
            "Logical": "ANY",
            "Conditions": [
                {
                    "LogicalOperator": "EQUALS",
                    "CrawlerName": crawler_name,
                    "CrawlState": "SUCCEEDED",
                }
            ],
        },
    }
    client.create_trigger(Name=trigger_name, **trigger_attributes)

    trigger = client.get_trigger(Name=trigger_name)["Trigger"]

    assert trigger["Name"] == trigger_name
    assert trigger["Type"] == "CONDITIONAL"
    assert trigger["State"] == "ACTIVATED"
    assert trigger["Actions"] == [{"JobName": job_name}]
    assert "Predicate" in trigger


def create_test_trigger(client, tags=None):
    job_name = create_test_job(client)
    trigger_name = str(uuid4())
    client.create_trigger(
        Name=trigger_name,
        Type="ON_DEMAND",
        Actions=[
            {
                "JobName": job_name,
            }
        ],
        Tags=tags or {},
    )
    return trigger_name


@mock_aws
def test_get_triggers_trigger_name_exists():
    client = create_glue_client()
    trigger_name = create_test_trigger(client)
    response = client.get_triggers()
    assert len(response["Triggers"]) == 1
    assert response["Triggers"][0]["Name"] == trigger_name


@mock_aws
def test_get_triggers_dependent_job_name():
    client = create_glue_client()

    create_test_trigger(client)
    job_name = create_test_job(client)
    trigger_name = str(uuid4())
    response = client.create_trigger(
        Name=trigger_name,
        Type="ON_DEMAND",
        Actions=[
            {
                "JobName": job_name,
            }
        ],
    )

    response = client.get_triggers(DependentJobName=job_name)
    assert len(response["Triggers"]) == 1
    assert response["Triggers"][0]["Name"] == trigger_name


@mock_aws
def test_start_trigger():
    client = create_glue_client()
    job_name = create_test_job(client)
    trigger_name = str(uuid4())
    trigger_attributes = {
        "Type": "SCHEDULED",
        "Schedule": "cron(5 3 * * ? *)",
        "Actions": [
            {
                "JobName": job_name,
            }
        ],
    }
    client.create_trigger(Name=trigger_name, **trigger_attributes)

    trigger = client.get_trigger(Name=trigger_name)["Trigger"]

    assert trigger["State"] == "CREATED"

    client.start_trigger(Name=trigger_name)

    trigger = client.get_trigger(Name=trigger_name)["Trigger"]

    assert trigger["State"] == "ACTIVATED"


@mock_aws
def test_stop_trigger():
    client = create_glue_client()
    job_name = create_test_job(client)
    trigger_name = str(uuid4())
    trigger_attributes = {
        "Type": "SCHEDULED",
        "Schedule": "cron(5 3 * * ? *)",
        "Actions": [
            {
                "JobName": job_name,
            }
        ],
        "StartOnCreation": True,
    }
    client.create_trigger(Name=trigger_name, **trigger_attributes)

    trigger = client.get_trigger(Name=trigger_name)["Trigger"]

    assert trigger["State"] == "ACTIVATED"

    client.stop_trigger(Name=trigger_name)

    trigger = client.get_trigger(Name=trigger_name)["Trigger"]

    assert trigger["State"] == "DEACTIVATED"


@mock_aws
def test_list_triggers():
    client = create_glue_client()
    trigger_name = create_test_trigger(client)
    response = client.list_triggers()
    assert response["TriggerNames"] == [trigger_name]
    assert "NextToken" not in response


@mock_aws
def test_list_triggers_dependent_job_name():
    client = create_glue_client()

    job_name = create_test_job(client)
    trigger_name = str(uuid4())
    trigger_attributes = {
        "Type": "ON_DEMAND",
        "Actions": [
            {
                "JobName": job_name,
            }
        ],
    }

    client.create_trigger(Name=trigger_name, **trigger_attributes)
    create_test_trigger(client)

    response = client.list_triggers()
    assert len(response["TriggerNames"]) == 2

    response = client.list_triggers(DependentJobName=job_name)
    assert len(response["TriggerNames"]) == 1
    assert response["TriggerNames"] == [trigger_name]


@mock_aws
def test_list_triggers_tags():
    client = create_glue_client()

    job_name = create_test_job(client)
    trigger_name = str(uuid4())
    trigger_attributes = {
        "Type": "ON_DEMAND",
        "Actions": [
            {
                "JobName": job_name,
            }
        ],
        "Tags": {
            "CreatedBy": "moto",
        },
    }

    client.create_trigger(Name=trigger_name, **trigger_attributes)
    create_test_trigger(client)

    response = client.list_triggers()
    assert len(response["TriggerNames"]) == 2

    response = client.list_triggers(Tags={"CreatedBy": "moto"})
    assert len(response["TriggerNames"]) == 1
    assert response["TriggerNames"] == [trigger_name]


@mock_aws
def test_batch_get_triggers():
    client = create_glue_client()
    trigger_name = create_test_trigger(client)

    response = client.batch_get_triggers(
        TriggerNames=[trigger_name, "trigger-not-found"]
    )

    assert len(response["Triggers"]) == 1
    assert len(response["TriggersNotFound"]) == 1


@mock_aws
def test_delete_trigger():
    client = create_glue_client()
    trigger_name = create_test_trigger(client)

    client.get_trigger(Name=trigger_name)

    client.delete_trigger(Name=trigger_name)

    with pytest.raises(ClientError) as exc:
        client.get_trigger(Name=trigger_name)

    assert exc.value.response["Error"]["Code"] == "EntityNotFoundException"


def create_test_session(client):
    session_id = str(uuid4())
    client.create_session(
        Id=session_id,
        Description="string",
        Role="arn_of_a_testing_role",
        Command={"Name": "string", "PythonVersion": "string"},
        Timeout=123,
        IdleTimeout=123,
        DefaultArguments={"string": "string"},
        Connections={
            "Connections": [
                "string",
            ]
        },
        MaxCapacity=123.0,
        NumberOfWorkers=123,
        WorkerType="Standard",
        SecurityConfiguration="string",
        GlueVersion="string",
        Tags={"string": "string"},
        RequestOrigin="string",
    )
    return session_id


@mock_aws
def test_create_session():
    client = create_glue_client()
    session_id = create_test_session(client)

    resp = client.get_session(Id=session_id)
    assert resp["Session"]["Id"] == session_id


@mock_aws
def test_get_session():
    client = create_glue_client()
    session_id = create_test_session(client)

    resp = client.get_session(Id=session_id)
    assert resp["Session"]["Id"] == session_id


@mock_aws
def test_list_sessions():
    client = create_glue_client()
    session_id = create_test_session(client)

    resp = client.list_sessions()
    assert session_id in resp["Ids"]


@mock_aws
def test_delete_session():
    client = create_glue_client()
    session_id = create_test_session(client)

    resp = client.delete_session(Id=session_id)
    assert resp["Id"] == session_id

    resp = client.list_sessions()
    assert session_id not in resp["Ids"]


@mock_aws
def test_stop_session():
    client = create_glue_client()
    session_id = create_test_session(client)

    resp = client.get_session(Id=session_id)
    assert resp["Session"]["Status"] == "READY"

    resp = client.stop_session(Id=session_id)
    assert resp["Id"] == session_id

    resp = client.get_session(Id=session_id)
    assert resp["Session"]["Status"] == "STOPPING"



import boto3
import pytest
from botocore.exceptions import ClientError

from moto import mock_aws
from moto.core import DEFAULT_ACCOUNT_ID as ACCOUNT_ID

from .test_helper_functions import CREATE_WEB_ACL_BODY, LIST_WEB_ACL_BODY


@mock_aws
@pytest.mark.parametrize(
    "region,partition",
    [("us-east-1", "aws"), ("cn-north-1", "aws-cn"), ("us-gov-east-1", "aws-us-gov")],
)
def test_create_web_acl(region, partition):
    conn = boto3.client("wafv2", region_name=region)
    res = conn.create_web_acl(**CREATE_WEB_ACL_BODY("John", "REGIONAL"))
    web_acl = res["Summary"]
    assert web_acl.get("Name") == "John"
    assert web_acl.get("ARN").startswith(
        f"arn:{partition}:wafv2:{region}:{ACCOUNT_ID}:regional/webacl/John/"
    )
    # Duplicate name - should raise error
    with pytest.raises(ClientError) as ex:
        conn.create_web_acl(**CREATE_WEB_ACL_BODY("John", "REGIONAL"))
    err = ex.value.response["Error"]
    assert (
        "AWS WAF could not perform the operation because some resource "
        "in your request is a duplicate of an existing one."
    ) in err["Message"]
    assert err["Code"] == "WafV2DuplicateItem"

    res = conn.create_web_acl(**CREATE_WEB_ACL_BODY("Carl", "CLOUDFRONT"))
    web_acl = res["Summary"]
    assert web_acl.get("ARN").startswith(
        f"arn:aws:wafv2:global:{ACCOUNT_ID}:global/webacl/Carl/"
    )


@mock_aws
def test_create_web_acl_with_all_arguments():
    client = boto3.client("wafv2", region_name="us-east-2")
    web_acl_id = client.create_web_acl(
        Name="test",
        Scope="CLOUDFRONT",
        DefaultAction={"Allow": {}},
        Description="test desc",
        VisibilityConfig={
            "SampledRequestsEnabled": False,
            "CloudWatchMetricsEnabled": False,
            "MetricName": "idk",
        },
        Rules=[
            {
                "Action": {"Allow": {}},
                "Name": "tf-acc-test-8205974093017792151-2",
                "Priority": 10,
                "Statement": {"GeoMatchStatement": {"CountryCodes": ["US", "NL"]}},
                "VisibilityConfig": {
                    "CloudWatchMetricsEnabled": False,
                    "MetricName": "tf-acc-test-8205974093017792151-2",
                    "SampledRequestsEnabled": False,
                },
            },
            {
                "Action": {"Count": {}},
                "Name": "tf-acc-test-8205974093017792151-1",
                "Priority": 5,
                "Statement": {
                    "SizeConstraintStatement": {
                        "ComparisonOperator": "LT",
                        "FieldToMatch": {"QueryString": {}},
                        "Size": 50,
                        "TextTransformations": [
                            {"Priority": 2, "Type": "CMD_LINE"},
                            {"Priority": 5, "Type": "NONE"},
                        ],
                    }
                },
                "VisibilityConfig": {
                    "CloudWatchMetricsEnabled": False,
                    "MetricName": "tf-acc-test-8205974093017792151-1",
                    "SampledRequestsEnabled": False,
                },
            },
        ],
    )["Summary"]["Id"]

    wacl = client.get_web_acl(Name="test", Scope="CLOUDFRONT", Id=web_acl_id)["WebACL"]
    assert wacl["Description"] == "test desc"
    assert wacl["DefaultAction"] == {"Allow": {}}
    assert wacl["VisibilityConfig"] == {
        "SampledRequestsEnabled": False,
        "CloudWatchMetricsEnabled": False,
        "MetricName": "idk",
    }
    assert len(wacl["Rules"]) == 2


@mock_aws
def test_get_web_acl():
    conn = boto3.client("wafv2", region_name="us-east-1")
    body = CREATE_WEB_ACL_BODY("John", "REGIONAL")
    web_acl_id = conn.create_web_acl(**body)["Summary"]["Id"]
    wacl = conn.get_web_acl(Name="John", Scope="REGIONAL", Id=web_acl_id)["WebACL"]
    assert wacl["Name"] == "John"
    assert wacl["Id"] == web_acl_id


@mock_aws
def test_list_web_acl():
    conn = boto3.client("wafv2", region_name="us-east-1")
    conn.create_web_acl(**CREATE_WEB_ACL_BODY("Daphne", "REGIONAL"))
    conn.create_web_acl(**CREATE_WEB_ACL_BODY("Penelope", "CLOUDFRONT"))
    conn.create_web_acl(**CREATE_WEB_ACL_BODY("Sarah", "REGIONAL"))
    res = conn.list_web_acls(**LIST_WEB_ACL_BODY("REGIONAL"))
    web_acls = res["WebACLs"]
    assert len(web_acls) == 2
    assert web_acls[0]["Name"] == "Daphne"
    assert web_acls[1]["Name"] == "Sarah"

    res = conn.list_web_acls(**LIST_WEB_ACL_BODY("CLOUDFRONT"))
    web_acls = res["WebACLs"]
    assert len(web_acls) == 1
    assert web_acls[0]["Name"] == "Penelope"


@mock_aws
def test_delete_web_acl():
    conn = boto3.client("wafv2", region_name="us-east-1")
    wacl_id = conn.create_web_acl(**CREATE_WEB_ACL_BODY("Daphne", "REGIONAL"))[
        "Summary"
    ]["Id"]

    conn.delete_web_acl(Name="Daphne", Id=wacl_id, Scope="REGIONAL", LockToken="n/a")

    res = conn.list_web_acls(**LIST_WEB_ACL_BODY("REGIONAL"))
    assert len(res["WebACLs"]) == 0

    with pytest.raises(ClientError) as exc:
        conn.get_web_acl(Name="Daphne", Scope="REGIONAL", Id=wacl_id)
    err = exc.value.response["Error"]
    assert err["Code"] == "WAFNonexistentItemException"
    assert err["Message"] == (
        "AWS WAF couldn’t perform the operation because your resource doesn’t exist."
    )


@mock_aws
def test_update_web_acl():
    conn = boto3.client("wafv2", region_name="us-east-1")
    wacl_id = conn.create_web_acl(**CREATE_WEB_ACL_BODY("Daphne", "REGIONAL"))[
        "Summary"
    ]["Id"]

    resp = conn.update_web_acl(
        Name="Daphne",
        Scope="REGIONAL",
        Id=wacl_id,
        DefaultAction={"Block": {"CustomResponse": {"ResponseCode": 412}}},
        Description="updated_desc",
        Rules=[
            {
                "Name": "rule1",
                "Priority": 456,
                "Statement": {},
                "VisibilityConfig": {
                    "SampledRequestsEnabled": True,
                    "CloudWatchMetricsEnabled": True,
                    "MetricName": "updated",
                },
            }
        ],
        LockToken="n/a",
        VisibilityConfig={
            "SampledRequestsEnabled": True,
            "CloudWatchMetricsEnabled": True,
            "MetricName": "updated",
        },
    )
    assert "NextLockToken" in resp

    acl = conn.get_web_acl(Name="Daphne", Scope="REGIONAL", Id=wacl_id)["WebACL"]
    assert acl["Description"] == "updated_desc"
    assert acl["DefaultAction"] == {"Block": {"CustomResponse": {"ResponseCode": 412}}}
    assert acl["Rules"] == [
        {
            "Name": "rule1",
            "Priority": 456,
            "Statement": {},
            "VisibilityConfig": {
                "SampledRequestsEnabled": True,
                "CloudWatchMetricsEnabled": True,
                "MetricName": "updated",
            },
        }
    ]
    assert acl["VisibilityConfig"] == {
        "SampledRequestsEnabled": True,
        "CloudWatchMetricsEnabled": True,
        "MetricName": "updated",
    }


@mock_aws
def test_ip_set_crud():
    client = boto3.client("wafv2", region_name="us-east-1")

    create_response = client.create_ip_set(
        Name="test-ip-set",
        Scope="CLOUDFRONT",
        Description="Test IP set",
        IPAddressVersion="IPV4",
        Addresses=["192.168.0.1/32", "10.0.0.0/8"],
        Tags=[{"Key": "Environment", "Value": "Test"}],
    )

    assert "Summary" in create_response
    summary = create_response["Summary"]

    assert summary["Id"]
    assert summary["LockToken"]
    assert summary["ARN"]
    assert summary["Name"] == "test-ip-set"
    assert "Test IP set" in summary["Description"]

    get_response = client.get_ip_set(
        Name=summary["Name"], Scope="CLOUDFRONT", Id=summary["Id"]
    )
    assert "IPSet" in get_response
    assert "LockToken" in get_response

    ip_set = get_response["IPSet"]
    assert ip_set["IPAddressVersion"] == "IPV4"

    with pytest.raises(ClientError) as e:
        client.update_ip_set(
            Name="test-ip-set",
            Scope="CLOUDFRONT",
            Id=summary["Id"],
            Addresses=["10.0.0.0/8"],
            LockToken="aaaaaaaaaaaaaaaaaa",  # invalid lock token
        )
    e.value.response["Error"]["Code"] == "WAFOptimisticLockException"
    e.value.response["Error"][
        "Message"
    ] == "AWS WAF couldn’t save your changes because someone changed the resource after you started to edit it. Reapply your changes."

    update_response = client.update_ip_set(
        Name="test-ip-set",
        Scope="CLOUDFRONT",
        Id=summary["Id"],
        Description="Updated test IP set",
        Addresses=["192.168.1.0/24", "10.0.0.0/8"],
        LockToken=get_response["LockToken"],
    )

    assert "NextLockToken" in update_response

    updated_get_response = client.get_ip_set(
        Name=summary["Name"], Scope="CLOUDFRONT", Id=summary["Id"]
    )

    updated_ip_set = updated_get_response["IPSet"]
    assert updated_ip_set["Description"] == "Updated test IP set"
    assert updated_get_response["LockToken"] == update_response["NextLockToken"]
    assert all(
        addr in ["192.168.1.0/24", "10.0.0.0/8"] for addr in updated_ip_set["Addresses"]
    )

    list_response = client.list_ip_sets(Scope="CLOUDFRONT")
    assert len(list_response["IPSets"]) == 1

    assert all(
        [
            key in list_response["IPSets"][0]
            for key in ["ARN", "Description", "Id", "LockToken", "Name"]
        ]
    )
    assert "NextMarker" in list_response

    client.delete_ip_set(
        Name=summary["Name"],
        Scope="CLOUDFRONT",
        Id=summary["Id"],
        LockToken=updated_get_response["LockToken"],
    )

    with pytest.raises(ClientError) as e:
        client.get_ip_set(Name=summary["Name"], Scope="CLOUDFRONT", Id=summary["Id"])
    e.value.response["Error"]["Code"] == "WAFNonexistentItemException"
    e.value.response["Error"][
        "Message"
    ] == "AWS WAF couldn’t perform the operation because your resource doesn’t exist."


@mock_aws
def test_logging_configuration_crud():
    wafv2_client = boto3.client("wafv2", region_name="us-east-1")
    create_web_acl_response = wafv2_client.create_web_acl(
        Name="TestWebACL",
        Scope="REGIONAL",
        DefaultAction={"Allow": {}},
        VisibilityConfig={
            "SampledRequestsEnabled": True,
            "CloudWatchMetricsEnabled": True,
            "MetricName": "TestWebACLMetric",
        },
        Rules=[],
    )
    web_acl_arn = create_web_acl_response["Summary"]["ARN"]

    # Create log groups
    logs_client = boto3.client("logs", region_name="us-east-1")
    logs_client.create_log_group(logGroupName="aws-waf-logs-test")
    log_group = logs_client.describe_log_groups(logGroupNamePrefix="aws-waf-logs-test")[
        "logGroups"
    ][0]

    create_response = wafv2_client.put_logging_configuration(
        LoggingConfiguration={
            "ResourceArn": web_acl_arn,
            "LogDestinationConfigs": [log_group["arn"]],
        }
    )

    assert "LoggingConfiguration" in create_response
    logging_configuration = create_response["LoggingConfiguration"]
    assert logging_configuration["ResourceArn"]

    get_response = wafv2_client.get_logging_configuration(
        ResourceArn=logging_configuration["ResourceArn"]
    )
    assert "LoggingConfiguration" in get_response

    list_response = wafv2_client.list_logging_configurations(
        Scope="REGIONAL",
    )
    assert len(list_response["LoggingConfigurations"]) > 0

    wafv2_client.delete_logging_configuration(
        ResourceArn=logging_configuration["ResourceArn"],
    )

    with pytest.raises(ClientError) as e:
        wafv2_client.get_logging_configuration(
            ResourceArn=logging_configuration["ResourceArn"]
        )
    e.value.response["Error"]["Code"] == "WAFNonexistentItemException"



"""directconnect base URL and path."""

from .responses import DirectConnectResponse

url_bases = [
    r"https?://directconnect\.(.+)\.amazonaws\.com",
]

url_paths = {
    "{0}/$": DirectConnectResponse.dispatch,
}



import abc
from typing import Any, Optional

from moto.stepfunctions.parser.asl.component.eval_component import EvalComponent
from moto.stepfunctions.parser.asl.eval.environment import Environment


class FunctionArgument(EvalComponent, abc.ABC):
    _value: Optional[Any]

    def __init__(self, value: Any = None):
        self._value = value

    def _eval_body(self, env: Environment) -> None:
        env.stack.append(self._value)



import json
from time import sleep

import boto3
import pytest
from botocore.exceptions import ClientError

from tests import allow_aws_request
from tests.test_dynamodb import dynamodb_aws_verified


def get_policy_doc(action):
    sts = boto3.client("sts", "us-east-1")
    account_id = sts.get_caller_identity()["Account"]

    policy_doc = {
        "Version": "2012-10-17",
        "Statement": [
            {
                "Principal": {"AWS": account_id},
                "Effect": "Allow",
                "Action": [action],
                "Resource": "*",
            }
        ],
    }
    return json.dumps(policy_doc)


@dynamodb_aws_verified()
@pytest.mark.aws_verified
def test_put_resource_policy(table_name=None):
    ddb = boto3.client("dynamodb", "us-east-1")

    table_arn = ddb.describe_table(TableName=table_name)["Table"]["TableArn"]
    revision_id = put_resource_policy(ddb, table_arn)

    policy = ddb.get_resource_policy(ResourceArn=table_arn)

    assert policy["RevisionId"] == revision_id
    assert "dynamodb:*" in json.loads(policy["Policy"])["Statement"][0]["Action"]


@dynamodb_aws_verified()
@pytest.mark.aws_verified
def test_put_resource_policy_multiple_times(table_name=None):
    ddb = boto3.client("dynamodb", "us-east-1")

    table_arn = ddb.describe_table(TableName=table_name)["Table"]["TableArn"]
    revision_id1 = put_resource_policy(ddb, table_arn)

    # Putting the same resource policy is a NO-OP
    revision_id2 = put_resource_policy(ddb, table_arn)
    assert revision_id1 == revision_id2

    # Only a different policy resorts in a different revision
    revision_id3 = put_resource_policy(ddb, table_arn, action="dynamodb:PutItem")
    assert revision_id3 != revision_id2
    # Because revision ID's are timestamps, they are comparable
    assert revision_id3 > revision_id2


@dynamodb_aws_verified()
@pytest.mark.aws_verified
def test_put_resource_policy_against_current_revision(table_name=None):
    ddb = boto3.client("dynamodb", "us-east-1")

    table_arn = ddb.describe_table(TableName=table_name)["Table"]["TableArn"]
    revision_id1 = put_resource_policy(ddb, table_arn)

    revision_id2 = put_resource_policy(
        ddb, table_arn, action="dynamodb:GetItem", revision_id=revision_id1
    )

    policy = ddb.get_resource_policy(ResourceArn=table_arn)

    assert policy["RevisionId"] == revision_id2
    assert "dynamodb:GetItem" in json.loads(policy["Policy"])["Statement"][0]["Action"]


@dynamodb_aws_verified()
@pytest.mark.aws_verified
def test_put_resource_policy_against_previous_revision(table_name=None):
    ddb = boto3.client("dynamodb", "us-east-1")

    table_arn = ddb.describe_table(TableName=table_name)["Table"]["TableArn"]
    revision_id1 = put_resource_policy(ddb, table_arn)

    revision_id2 = put_resource_policy(
        ddb, table_arn, action="dynamodb:PutItem", revision_id=revision_id1
    )

    with pytest.raises(ClientError) as exc:
        put_resource_policy(
            ddb, table_arn, action="dynamodb:GetItem", revision_id=revision_id1
        )
    err = exc.value.response["Error"]
    assert err["Code"] == "PolicyNotFoundException"
    assert (
        err["Message"]
        == f"Resource-based policy not found for the provided ResourceArn: Requested update for policy with revision id {revision_id1}, but the policy associated to target table {table_name} has revision id {revision_id2}."
    )

    assert ddb.get_resource_policy(ResourceArn=table_arn)["RevisionId"] == revision_id2


@dynamodb_aws_verified()
@pytest.mark.aws_verified
def test_put_resource_policy_against_norevision(table_name=None):
    ddb = boto3.client("dynamodb", "us-east-1")

    table_arn = ddb.describe_table(TableName=table_name)["Table"]["TableArn"]
    revision_id = put_resource_policy(ddb, table_arn, revision_id="NO_POLICY")

    policy = ddb.get_resource_policy(ResourceArn=table_arn)

    assert policy["RevisionId"] == revision_id


@dynamodb_aws_verified()
@pytest.mark.aws_verified
def test_put_resource_policy_against_norevision_when_revision_exists(table_name=None):
    ddb = boto3.client("dynamodb", "us-east-1")

    table_arn = ddb.describe_table(TableName=table_name)["Table"]["TableArn"]
    revision_id = put_resource_policy(ddb, table_arn)

    with pytest.raises(ClientError) as exc:
        put_resource_policy(ddb, table_arn, revision_id="NO_POLICY")
    err = exc.value.response["Error"]
    assert err["Code"] == "PolicyNotFoundException"
    assert (
        err["Message"]
        == f"Resource-based policy not found for the provided ResourceArn: Requested policy update expecting none to be present for table {table_name}, but a policy exists with revision id {revision_id}."
    )

    assert ddb.get_resource_policy(ResourceArn=table_arn)["RevisionId"] == revision_id


@dynamodb_aws_verified()
@pytest.mark.aws_verified
def test_get_resource_policy_from_resource_without(table_name=None):
    ddb = boto3.client("dynamodb", "us-east-1")

    table_arn = ddb.describe_table(TableName=table_name)["Table"]["TableArn"]
    with pytest.raises(ClientError) as exc:
        ddb.get_resource_policy(ResourceArn=table_arn)
    err = exc.value.response["Error"]
    assert err["Code"] == "PolicyNotFoundException"
    assert (
        err["Message"]
        == f"Resource-based policy not found for the provided ResourceArn: {table_arn}"
    )


@dynamodb_aws_verified()
@pytest.mark.aws_verified
def test_put_resource_policy_unknown_resource(table_name=None):
    ddb = boto3.client("dynamodb", "us-east-1")

    table_arn = ddb.describe_table(TableName=table_name)["Table"]["TableArn"]
    with pytest.raises(ClientError) as exc:
        put_resource_policy(ddb, table_arn + "_")
    err = exc.value.response["Error"]
    assert err["Code"] == "ResourceNotFoundException"
    assert (
        err["Message"]
        == f"Requested resource not found: Table: {table_name}_ not found"
    )


@dynamodb_aws_verified()
@pytest.mark.aws_verified
def test_get_resource_policy_from_unknown_resource(table_name=None):
    ddb = boto3.client("dynamodb", "us-east-1")

    table_arn = ddb.describe_table(TableName=table_name)["Table"]["TableArn"]
    with pytest.raises(ClientError) as exc:
        ddb.get_resource_policy(ResourceArn=table_arn + "unknown")
    err = exc.value.response["Error"]
    assert err["Code"] == "ResourceNotFoundException"
    assert (
        err["Message"]
        == f"Requested resource not found: Table: {table_name}unknown not found"
    )


@dynamodb_aws_verified()
@pytest.mark.aws_verified
def test_delete_resource_policy_from_unknown_resource(table_name=None):
    ddb = boto3.client("dynamodb", "us-east-1")

    table_arn = ddb.describe_table(TableName=table_name)["Table"]["TableArn"]
    with pytest.raises(ClientError) as exc:
        ddb.delete_resource_policy(ResourceArn=table_arn + "unknown")
    err = exc.value.response["Error"]
    assert err["Code"] == "ResourceNotFoundException"
    assert (
        err["Message"]
        == f"Requested resource not found: Table: {table_name}unknown not found"
    )


@dynamodb_aws_verified()
@pytest.mark.aws_verified
def test_delete_resource_policy_from_unknown_revision(table_name=None):
    ddb = boto3.client("dynamodb", "us-east-1")

    table_arn = ddb.describe_table(TableName=table_name)["Table"]["TableArn"]
    revision = put_resource_policy(ddb, table_arn)

    with pytest.raises(ClientError) as exc:
        ddb.delete_resource_policy(ResourceArn=table_arn, ExpectedRevisionId="unknown")
    err = exc.value.response["Error"]
    assert err["Code"] == "PolicyNotFoundException"
    assert (
        err["Message"]
        == f"Resource-based policy not found for the provided ResourceArn: Requested update for policy with revision id unknown, but the policy associated to target table {table_name} has revision id {revision}."
    )


@dynamodb_aws_verified()
@pytest.mark.aws_verified
def test_delete_resource_policy_from_former_revision(table_name=None):
    ddb = boto3.client("dynamodb", "us-east-1")

    table_arn = ddb.describe_table(TableName=table_name)["Table"]["TableArn"]
    old_revision_id = put_resource_policy(ddb, table_arn)

    new_revision_id = put_resource_policy(ddb, table_arn, action="dynamodb:PutItem")

    with pytest.raises(ClientError) as exc:
        ddb.delete_resource_policy(
            ResourceArn=table_arn, ExpectedRevisionId=old_revision_id
        )
    err = exc.value.response["Error"]
    assert err["Code"] == "PolicyNotFoundException"
    assert (
        err["Message"]
        == f"Resource-based policy not found for the provided ResourceArn: Requested update for policy with revision id {old_revision_id}, but the policy associated to target table {table_name} has revision id {new_revision_id}."
    )


@dynamodb_aws_verified()
@pytest.mark.aws_verified
def test_delete_resource_policy(table_name=None):
    ddb = boto3.client("dynamodb", "us-east-1")

    table_arn = ddb.describe_table(TableName=table_name)["Table"]["TableArn"]
    put_resource_policy(ddb, table_arn)

    ddb.delete_resource_policy(ResourceArn=table_arn)
    if allow_aws_request():
        # Wait for deletion to finish
        sleep(15)

    with pytest.raises(ClientError) as exc:
        ddb.get_resource_policy(ResourceArn=table_arn)
    assert exc.value.response["Error"]["Code"] == "PolicyNotFoundException"


@dynamodb_aws_verified()
@pytest.mark.aws_verified
def test_delete_resource_policy_with_revision(table_name=None):
    ddb = boto3.client("dynamodb", "us-east-1")

    table_arn = ddb.describe_table(TableName=table_name)["Table"]["TableArn"]
    old_revision_id = put_resource_policy(ddb, table_arn)

    ddb.delete_resource_policy(
        ResourceArn=table_arn, ExpectedRevisionId=old_revision_id
    )
    if allow_aws_request():
        # Wait for deletion to finish
        sleep(15)

    with pytest.raises(ClientError) as exc:
        ddb.get_resource_policy(ResourceArn=table_arn)
    assert exc.value.response["Error"]["Code"] == "PolicyNotFoundException"


def put_resource_policy(ddb, table_arn, action="dynamodb:*", revision_id=None):
    kwargs = {}
    if revision_id:
        kwargs["ExpectedRevisionId"] = revision_id
    policy_doc = get_policy_doc(action)
    revision_id = ddb.put_resource_policy(
        ResourceArn=table_arn, Policy=policy_doc, **kwargs
    )["RevisionId"]
    # Put Resource Policy is an async action
    # Periodically poll to see when the policy is ready
    for x in range(20):
        try:
            policy = ddb.get_resource_policy(ResourceArn=table_arn)
            assert policy["RevisionId"] == revision_id
            if allow_aws_request():
                # Policy is supposedly ready-  but may still complain that a ResourcePolicy update is in progress
                # Hopefully AWS is ready afterwards
                sleep(1)
            return revision_id
        except (ClientError, AssertionError):
            # Policy is not ready yet
            sleep(1)
    assert False, f"Revision {revision_id} wasn't ready in 20 seconds"



from random import randint
from unittest import SkipTest

import boto3
import pytest
from botocore.exceptions import ClientError

from moto import mock_aws, settings
from moto.core import DEFAULT_ACCOUNT_ID as OWNER_ID


@mock_aws
def test_default_network_acl_created_with_vpc():
    client = boto3.client("ec2", region_name="us-east-1")
    ec2 = boto3.resource("ec2", region_name="us-east-1")
    vpc = ec2.create_vpc(CidrBlock="10.0.0.0/16")

    all_network_acls = client.describe_network_acls()["NetworkAcls"]
    our_acl = [a for a in all_network_acls if a["VpcId"] == vpc.id]
    assert len(our_acl) == 1
    assert our_acl[0]["IsDefault"] is True


@mock_aws
def test_network_create_and_list_acls():
    client = boto3.client("ec2", region_name="us-east-1")
    ec2 = boto3.resource("ec2", region_name="us-east-1")
    vpc = ec2.create_vpc(CidrBlock="10.0.0.0/16")
    created_acl = ec2.create_network_acl(VpcId=vpc.id)

    all_network_acls = client.describe_network_acls()["NetworkAcls"]

    acl_found = [
        a for a in all_network_acls if a["NetworkAclId"] == created_acl.network_acl_id
    ][0]
    assert acl_found["VpcId"] == vpc.id
    assert acl_found["Tags"] == []
    assert acl_found["IsDefault"] is False


@mock_aws
def test_new_subnet_associates_with_default_network_acl():
    if settings.TEST_SERVER_MODE:
        raise SkipTest("ServerMode will have conflicting CidrBlocks")
    client = boto3.client("ec2", region_name="us-east-1")
    ec2 = boto3.resource("ec2", region_name="us-east-1")
    default_vpc = client.describe_vpcs()["Vpcs"][0]

    subnet = ec2.create_subnet(VpcId=default_vpc["VpcId"], CidrBlock="172.31.112.0/20")
    all_network_acls = client.describe_network_acls()["NetworkAcls"]
    assert len(all_network_acls) == 1

    acl = all_network_acls[0]
    assert len(acl["Associations"]) == 7
    assert subnet.id in [a["SubnetId"] for a in acl["Associations"]]


@mock_aws
def test_network_acl_entries():
    client = boto3.client("ec2", region_name="us-east-1")
    ec2 = boto3.resource("ec2", region_name="us-east-1")
    vpc = ec2.create_vpc(CidrBlock="10.0.0.0/16")

    network_acl = ec2.create_network_acl(VpcId=vpc.id)

    client.create_network_acl_entry(
        NetworkAclId=network_acl.id,
        RuleNumber=110,
        Protocol="6",  # TCP
        RuleAction="ALLOW",
        CidrBlock="0.0.0.0/0",
        Egress=False,
        PortRange={"From": 443, "To": 443},
    )

    all_network_acls = client.describe_network_acls()["NetworkAcls"]

    test_network_acl = next(
        na for na in all_network_acls if na["NetworkAclId"] == network_acl.id
    )
    assert test_network_acl["IsDefault"] is False

    entries = test_network_acl["Entries"]
    assert len(entries) == 1
    assert entries[0]["RuleNumber"] == 110
    assert entries[0]["Protocol"] == "6"
    assert entries[0]["RuleAction"] == "ALLOW"
    assert entries[0]["Egress"] is False
    assert entries[0]["PortRange"] == {"To": 443, "From": 443}
    assert entries[0]["CidrBlock"] == "0.0.0.0/0"


@mock_aws
def test_delete_network_acl_entry():
    client = boto3.client("ec2", region_name="us-east-1")
    ec2 = boto3.resource("ec2", region_name="us-east-1")
    vpc = ec2.create_vpc(CidrBlock="10.0.0.0/16")

    network_acl = ec2.create_network_acl(VpcId=vpc.id)

    client.create_network_acl_entry(
        NetworkAclId=network_acl.id,
        RuleNumber=110,
        Protocol="6",  # TCP
        RuleAction="ALLOW",
        CidrBlock="0.0.0.0/0",
        Egress=False,
        PortRange={"From": 443, "To": 443},
    )
    client.delete_network_acl_entry(
        NetworkAclId=network_acl.id, RuleNumber=110, Egress=False
    )

    all_network_acls = client.describe_network_acls()["NetworkAcls"]

    test_network_acl = next(
        na for na in all_network_acls if na["NetworkAclId"] == network_acl.id
    )
    assert len(test_network_acl["Entries"]) == 0


@mock_aws
def test_replace_network_acl_entry():
    client = boto3.client("ec2", region_name="us-east-1")
    ec2 = boto3.resource("ec2", region_name="us-east-1")
    vpc = ec2.create_vpc(CidrBlock="10.0.0.0/16")

    network_acl = ec2.create_network_acl(VpcId=vpc.id)

    client.create_network_acl_entry(
        NetworkAclId=network_acl.id,
        RuleNumber=110,
        Protocol="6",  # TCP
        RuleAction="ALLOW",
        CidrBlock="0.0.0.0/0",
        Egress=False,
        PortRange={"From": 443, "To": 443},
    )
    client.replace_network_acl_entry(
        NetworkAclId=network_acl.id,
        RuleNumber=110,
        Protocol="-1",
        RuleAction="DENY",
        CidrBlock="0.0.0.0/0",
        Egress=False,
        PortRange={"From": 22, "To": 22},
    )

    all_network_acls = client.describe_network_acls()["NetworkAcls"]

    test_network_acl = next(
        na for na in all_network_acls if na["NetworkAclId"] == network_acl.id
    )
    entries = test_network_acl["Entries"]
    assert len(entries) == 1
    assert entries[0]["RuleNumber"] == 110
    assert entries[0]["Protocol"] == "-1"
    assert entries[0]["RuleAction"] == "DENY"
    assert entries[0]["PortRange"] == {"To": 22, "From": 22}


@mock_aws
def test_delete_network_acl():
    client = boto3.client("ec2", region_name="us-east-1")
    ec2 = boto3.resource("ec2", region_name="us-east-1")
    vpc = ec2.create_vpc(CidrBlock="10.0.0.0/16")
    network_acl = ec2.create_network_acl(VpcId=vpc.id)

    all_network_acls = client.describe_network_acls()["NetworkAcls"]

    assert (
        any(acl["NetworkAclId"] == network_acl.id for acl in all_network_acls)
        is not None
    )

    client.delete_network_acl(NetworkAclId=network_acl.id)

    updated_network_acls = client.describe_network_acls()["NetworkAcls"]

    assert not any(
        acl["NetworkAclId"] == network_acl.id for acl in updated_network_acls
    )


@mock_aws
def test_network_acl_tagging():
    client = boto3.client("ec2", region_name="us-east-1")
    ec2 = boto3.resource("ec2", region_name="us-east-1")
    vpc = ec2.create_vpc(CidrBlock="10.0.0.0/16")
    network_acl = ec2.create_network_acl(VpcId=vpc.id)

    network_acl.create_tags(Tags=[{"Key": "a key", "Value": "some value"}])

    tag = client.describe_tags(
        Filters=[{"Name": "resource-id", "Values": [network_acl.id]}]
    )["Tags"][0]
    assert tag["ResourceId"] == network_acl.id
    assert tag["Key"] == "a key"
    assert tag["Value"] == "some value"

    all_network_acls = client.describe_network_acls()["NetworkAcls"]
    test_network_acl = next(
        na for na in all_network_acls if na["NetworkAclId"] == network_acl.id
    )
    assert test_network_acl["Tags"] == [{"Value": "some value", "Key": "a key"}]


@mock_aws
def test_new_subnet_in_new_vpc_associates_with_default_network_acl():
    ec2 = boto3.resource("ec2", region_name="us-west-1")
    new_vpc = ec2.create_vpc(CidrBlock="10.0.0.0/16")
    new_vpc.reload()

    subnet = ec2.create_subnet(VpcId=new_vpc.id, CidrBlock="10.0.0.0/24")
    subnet.reload()

    new_vpcs_default_network_acl = next(iter(new_vpc.network_acls.all()), None)
    new_vpcs_default_network_acl.reload()
    assert new_vpcs_default_network_acl.vpc_id == new_vpc.id
    assert len(new_vpcs_default_network_acl.associations) == 1
    assert new_vpcs_default_network_acl.associations[0]["SubnetId"] == subnet.id


@mock_aws
def test_default_network_acl_default_entries():
    ec2 = boto3.resource("ec2", region_name="us-west-1")
    client = boto3.client("ec2", region_name="us-west-1")

    if not settings.TEST_SERVER_MODE:
        # Can't know whether the first ACL is the default in ServerMode
        default_network_acl = next(iter(ec2.network_acls.all()), None)
        assert default_network_acl.is_default is True
        assert len(default_network_acl.entries) == 4

    vpc = client.create_vpc(CidrBlock="10.0.0.0/16")
    vpc_id = vpc["Vpc"]["VpcId"]

    default_acls = client.describe_network_acls(
        Filters=[
            {"Name": "default", "Values": ["true"]},
            {"Name": "vpc-id", "Values": [vpc_id]},
        ]
    )["NetworkAcls"]
    default_acl = default_acls[0]

    unique_entries = []
    for entry in default_acl["Entries"]:
        assert entry["CidrBlock"] == "0.0.0.0/0"
        assert entry["Protocol"] == "-1"
        assert entry["RuleNumber"] in [100, 32767]
        assert entry["RuleAction"] in ["allow", "deny"]
        assert isinstance(entry["Egress"], bool)
        if entry["RuleAction"] == "allow":
            assert entry["RuleNumber"] == 100
        else:
            assert entry["RuleNumber"] == 32767
        if entry not in unique_entries:
            unique_entries.append(entry)

    assert len(unique_entries) == 4


@mock_aws
def test_delete_default_network_acl_default_entry():
    if settings.TEST_SERVER_MODE:
        raise SkipTest(
            "Don't want to mess with default ACLs in ServerMode, as other tests may need them"
        )
    ec2 = boto3.resource("ec2", region_name="us-west-1")
    default_network_acl = next(iter(ec2.network_acls.all()), None)
    assert default_network_acl.is_default is True

    assert len(default_network_acl.entries) == 4
    first_default_network_acl_entry = default_network_acl.entries[0]

    default_network_acl.delete_entry(
        Egress=first_default_network_acl_entry["Egress"],
        RuleNumber=first_default_network_acl_entry["RuleNumber"],
    )

    assert len(default_network_acl.entries) == 3


@mock_aws
def test_duplicate_network_acl_entry():
    ec2 = boto3.resource("ec2", region_name="us-west-1")
    default_network_acl = next(iter(ec2.network_acls.all()), None)
    assert default_network_acl.is_default is True

    rule_number = randint(0, 9999)
    egress = True
    default_network_acl.create_entry(
        CidrBlock="0.0.0.0/0",
        Egress=egress,
        Protocol="-1",
        RuleAction="allow",
        RuleNumber=rule_number,
    )

    with pytest.raises(ClientError) as ex:
        default_network_acl.create_entry(
            CidrBlock="10.0.0.0/0",
            Egress=egress,
            Protocol="-1",
            RuleAction="deny",
            RuleNumber=rule_number,
        )
    assert (
        str(ex.value)
        == f"An error occurred (NetworkAclEntryAlreadyExists) when calling the CreateNetworkAclEntry operation: The network acl entry identified by {rule_number} already exists."
    )


@mock_aws
def test_describe_network_acls():
    conn = boto3.client("ec2", region_name="us-west-2")

    vpc = conn.create_vpc(CidrBlock="10.0.0.0/16")
    vpc_id = vpc["Vpc"]["VpcId"]

    network_acl = conn.create_network_acl(VpcId=vpc_id)

    network_acl_id = network_acl["NetworkAcl"]["NetworkAclId"]

    resp = conn.describe_network_acls(NetworkAclIds=[network_acl_id])
    result = resp["NetworkAcls"]

    assert len(result) == 1
    assert result[0]["NetworkAclId"] == network_acl_id

    resp2 = conn.describe_network_acls()["NetworkAcls"]
    assert network_acl_id in [na["NetworkAclId"] for na in resp2]

    resp3 = conn.describe_network_acls(
        Filters=[{"Name": "owner-id", "Values": [OWNER_ID]}]
    )["NetworkAcls"]
    assert network_acl_id in [na["NetworkAclId"] for na in resp3]

    # Assertions for filters
    network_acl_id = conn.create_network_acl(VpcId=vpc_id)["NetworkAcl"]["NetworkAclId"]
    cidr_block = "0.0.0.0/24"
    protocol = "17"  # UDP
    rule_number = 420
    rule_action = "allow"
    conn.create_network_acl_entry(
        NetworkAclId=network_acl_id,
        CidrBlock=cidr_block,
        Protocol=protocol,
        RuleNumber=rule_number,
        RuleAction=rule_action,
        Egress=False,
    )

    # Ensure filtering by entry CIDR block
    resp4 = conn.describe_network_acls(
        Filters=[{"Name": "entry.cidr", "Values": [cidr_block]}]
    )
    assert len(resp4["NetworkAcls"]) == 1
    assert resp4["NetworkAcls"][0]["NetworkAclId"] == network_acl_id
    assert cidr_block in [
        entry["CidrBlock"] for entry in resp4["NetworkAcls"][0]["Entries"]
    ]

    # Ensure filtering by entry protocol
    resp4 = conn.describe_network_acls(
        Filters=[{"Name": "entry.protocol", "Values": [protocol]}]
    )
    assert len(resp4["NetworkAcls"]) == 1
    assert resp4["NetworkAcls"][0]["NetworkAclId"] == network_acl_id
    assert protocol in [
        entry["Protocol"] for entry in resp4["NetworkAcls"][0]["Entries"]
    ]

    # Ensure filtering by entry rule number
    resp4 = conn.describe_network_acls(
        Filters=[{"Name": "entry.rule-number", "Values": [str(rule_number)]}]
    )
    assert len(resp4["NetworkAcls"]) == 1
    assert resp4["NetworkAcls"][0]["NetworkAclId"] == network_acl_id
    assert rule_number in [
        entry["RuleNumber"] for entry in resp4["NetworkAcls"][0]["Entries"]
    ]

    resp4 = conn.describe_network_acls(
        Filters=[{"Name": "entry.rule-number", "Values": [str(rule_number + 1)]}]
    )
    assert len(resp4["NetworkAcls"]) == 0

    # Ensure filtering by egress flag
    resp4 = conn.describe_network_acls(
        Filters=[{"Name": "entry.egress", "Values": ["false"]}]
    )
    assert network_acl_id in [entry["NetworkAclId"] for entry in resp4["NetworkAcls"]]
    # the ACL with network_acl_id contains no entries with Egress=True
    resp4 = conn.describe_network_acls(
        Filters=[{"Name": "entry.egress", "Values": ["true"]}]
    )
    assert network_acl_id not in [
        entry["NetworkAclId"] for entry in resp4["NetworkAcls"]
    ]

    # Ensure filtering by rule action
    resp4 = conn.describe_network_acls(
        Filters=[
            {"Name": "entry.rule-action", "Values": [rule_action]},
            {"Name": "id", "Values": [network_acl_id]},
        ]
    )
    assert len(resp4["NetworkAcls"]) == 1
    assert resp4["NetworkAcls"][0]["NetworkAclId"] == network_acl_id
    assert rule_action in [
        entry["RuleAction"] for entry in resp4["NetworkAcls"][0]["Entries"]
    ]

    with pytest.raises(ClientError) as ex:
        conn.describe_network_acls(NetworkAclIds=["1"])

    assert (
        str(ex.value)
        == "An error occurred (InvalidRouteTableID.NotFound) when calling the DescribeNetworkAcls operation: The routeTable ID '1' does not exist"
    )


@mock_aws
def test_describe_network_acls_associations_by_subnet():
    ec2_client = boto3.client("ec2", region_name="us-west-2")
    vpc = ec2_client.create_vpc(CidrBlock="10.0.0.0/16")
    vpc_id = vpc["Vpc"]["VpcId"]
    all_acls = ec2_client.describe_network_acls(
        Filters=[{"Name": "vpc-id", "Values": [vpc_id]}]
    )["NetworkAcls"]
    subnet_1_id = ec2_client.create_subnet(VpcId=vpc_id, CidrBlock="10.0.1.0/24")[
        "Subnet"
    ]["SubnetId"]
    subnet_2_id = ec2_client.create_subnet(VpcId=vpc_id, CidrBlock="10.0.2.0/24")[
        "Subnet"
    ]["SubnetId"]
    subnet_3_id = ec2_client.create_subnet(VpcId=vpc_id, CidrBlock="10.0.3.0/24")[
        "Subnet"
    ]["SubnetId"]
    default_acl_id = next(acl["NetworkAclId"] for acl in all_acls if acl["IsDefault"])
    custom_acl_id = ec2_client.create_network_acl(VpcId=vpc_id)["NetworkAcl"][
        "NetworkAclId"
    ]

    # Subnets should be associated with the default ACL by default
    assert_subnet_is_associated_to_acl(ec2_client, subnet_1_id, default_acl_id)
    assert_subnet_is_associated_to_acl(ec2_client, subnet_3_id, default_acl_id)
    assert_subnet_is_associated_to_acl(ec2_client, subnet_2_id, default_acl_id)

    associate_subnet_to_acl(ec2_client, subnet_1_id, custom_acl_id)
    associate_subnet_to_acl(ec2_client, subnet_3_id, custom_acl_id)
    assert_subnet_is_associated_to_acl(ec2_client, subnet_1_id, custom_acl_id)
    assert_subnet_is_associated_to_acl(ec2_client, subnet_3_id, custom_acl_id)
    assert_subnet_is_associated_to_acl(ec2_client, subnet_2_id, default_acl_id)

    # Re-associate Subnet 1 with the default ACL to test re-association
    associate_subnet_to_acl(ec2_client, subnet_1_id, default_acl_id)
    assert_subnet_is_associated_to_acl(ec2_client, subnet_1_id, default_acl_id)
    assert_subnet_is_associated_to_acl(ec2_client, subnet_2_id, default_acl_id)
    assert_subnet_is_associated_to_acl(ec2_client, subnet_3_id, custom_acl_id)

    # Attempting to associate a non-existent ACL with a subnet should raise an error
    with pytest.raises(ClientError):
        associate_subnet_to_acl(ec2_client, subnet_1_id, "non-existent-acl")

    # Attempting to delete a custom network ACL should fail if it has any subnets associated
    with pytest.raises(ClientError) as exc:
        ec2_client.delete_network_acl(NetworkAclId=custom_acl_id)
    err = exc.value.response["Error"]
    assert err["Code"] == "DependencyViolation"
    assert (
        err["Message"]
        == f"The network ACL '{custom_acl_id}' has dependencies and cannot be deleted."
    )

    # Re-associate Subnet 3 with the default ACL to test deletion of custom ACL
    associate_subnet_to_acl(ec2_client, subnet_3_id, default_acl_id)
    ec2_client.delete_network_acl(NetworkAclId=custom_acl_id)
    assert_subnet_is_associated_to_acl(ec2_client, subnet_3_id, default_acl_id)


def associate_subnet_to_acl(ec2_client, subnet_id: str, acl_id: str):
    assoc_id = get_subnet_acl_association_id(ec2_client, subnet_id)
    ec2_client.replace_network_acl_association(
        NetworkAclId=acl_id, AssociationId=assoc_id
    )


def get_subnet_acl_association_id(ec2_client, subnet_id: str):
    network_acls_description = ec2_client.describe_network_acls(
        Filters=[{"Name": "association.subnet-id", "Values": [subnet_id]}]
    )
    associations = network_acls_description["NetworkAcls"][0]["Associations"]
    return next(
        assoc["NetworkAclAssociationId"]
        for assoc in associations
        if assoc["SubnetId"] == subnet_id
    )


def assert_subnet_is_associated_to_acl(ec2_client, subnet_id: str, acl_id: str):
    network_acl_description = ec2_client.describe_network_acls(
        Filters=[{"Name": "association.subnet-id", "Values": [subnet_id]}]
    )
    acls_subnet = network_acl_description["NetworkAcls"]
    # There should be only one ACL associated with the subnet at any given time
    assert len(acls_subnet) == 1
    assert acls_subnet[0]["NetworkAclId"] == acl_id


@mock_aws
def test_create_network_acl_with_tags():
    conn = boto3.client("ec2", region_name="us-west-2")

    vpc = conn.create_vpc(CidrBlock="10.0.0.0/16")
    vpc_id = vpc["Vpc"]["VpcId"]

    network_acl = conn.create_network_acl(
        VpcId=vpc_id,
        TagSpecifications=[
            {
                "ResourceType": "network-acl",
                "Tags": [{"Key": "test", "Value": "TestTags"}],
            }
        ],
    )

    assert (len(network_acl.get("NetworkAcl").get("Tags"))) == 1
    assert network_acl.get("NetworkAcl").get("Tags") == [
        {"Key": "test", "Value": "TestTags"}
    ]



from uuid import uuid4

import boto3
import pytest
from botocore.exceptions import ClientError

from moto import mock_aws

DEFAULT_REGION_NAME = "us-east-1"


@mock_aws
def test_get_bucket_replication_for_unexisting_bucket():
    bucket_name = str(uuid4())
    s3_client = boto3.client("s3", region_name=DEFAULT_REGION_NAME)
    with pytest.raises(ClientError) as exc:
        s3_client.get_bucket_replication(Bucket=bucket_name)
    err = exc.value.response["Error"]
    assert err["Code"] == "NoSuchBucket"
    assert err["Message"] == "The specified bucket does not exist"
    assert err["BucketName"] == bucket_name


@mock_aws
def test_get_bucket_replication_bucket_without_replication():
    bucket_name = str(uuid4())
    s3_client = boto3.client("s3", region_name=DEFAULT_REGION_NAME)
    s3_client.create_bucket(Bucket=bucket_name)

    with pytest.raises(ClientError) as exc:
        s3_client.get_bucket_replication(Bucket=bucket_name)
    err = exc.value.response["Error"]
    assert err["Code"] == "ReplicationConfigurationNotFoundError"
    assert err["Message"] == "The replication configuration was not found"
    assert err["BucketName"] == bucket_name


@mock_aws
def test_delete_bucket_replication_unknown_bucket():
    bucket_name = str(uuid4())
    s3_client = boto3.client("s3", region_name=DEFAULT_REGION_NAME)
    with pytest.raises(ClientError) as exc:
        s3_client.delete_bucket_replication(Bucket=bucket_name)
    err = exc.value.response["Error"]
    assert err["Code"] == "NoSuchBucket"
    assert err["Message"] == "The specified bucket does not exist"
    assert err["BucketName"] == bucket_name


@mock_aws
def test_delete_bucket_replication_bucket_without_replication():
    bucket_name = str(uuid4())
    s3_client = boto3.client("s3", region_name=DEFAULT_REGION_NAME)

    s3_client.create_bucket(Bucket=bucket_name)
    # No-op
    s3_client.delete_bucket_replication(Bucket=bucket_name)


@mock_aws
def test_create_replication_without_versioning():
    bucket_name = str(uuid4())
    s3_client = boto3.client("s3", region_name=DEFAULT_REGION_NAME)
    s3_client.create_bucket(Bucket=bucket_name)

    with pytest.raises(ClientError) as exc:
        s3_client.put_bucket_replication(
            Bucket=bucket_name,
            ReplicationConfiguration={
                "Role": "myrole",
                "Rules": [
                    {"Destination": {"Bucket": "secondbucket"}, "Status": "Enabled"}
                ],
            },
        )
    err = exc.value.response["Error"]
    assert err["Code"] == "InvalidRequest"
    assert err["Message"] == (
        "Versioning must be 'Enabled' on the bucket to apply a replication configuration"
    )
    assert err["BucketName"] == bucket_name


@mock_aws
def test_create_and_retrieve_replication_with_single_rules():
    bucket_name = str(uuid4())
    s3_client = boto3.client("s3", region_name=DEFAULT_REGION_NAME)

    s3_client.create_bucket(Bucket=bucket_name)
    s3_client.put_bucket_versioning(
        Bucket=bucket_name, VersioningConfiguration={"Status": "Enabled"}
    )
    s3_client.put_bucket_replication(
        Bucket=bucket_name,
        ReplicationConfiguration={
            "Role": "myrole",
            "Rules": [
                {
                    "ID": "firstrule",
                    "Priority": 2,
                    "Destination": {"Bucket": "secondbucket"},
                    "Status": "Enabled",
                }
            ],
        },
    )

    config = s3_client.get_bucket_replication(Bucket=bucket_name)[
        "ReplicationConfiguration"
    ]
    assert config == {
        "Role": "myrole",
        "Rules": [
            {
                "DeleteMarkerReplication": {"Status": "Disabled"},
                "Destination": {"Bucket": "secondbucket"},
                "Filter": {"Prefix": ""},
                "ID": "firstrule",
                "Priority": 2,
                "Status": "Enabled",
            }
        ],
    }

    s3_client.delete_bucket_replication(Bucket=bucket_name)

    # Can't retrieve replication that has been deleted
    with pytest.raises(ClientError) as exc:
        s3_client.get_bucket_replication(Bucket=bucket_name)
    err = exc.value.response["Error"]
    assert err["Code"] == "ReplicationConfigurationNotFoundError"
    assert err["Message"] == "The replication configuration was not found"
    assert err["BucketName"] == bucket_name


@mock_aws
def test_create_and_retrieve_replication_with_multiple_rules():
    bucket_name = str(uuid4())
    s3_client = boto3.client("s3", region_name=DEFAULT_REGION_NAME)

    s3_client.create_bucket(Bucket=bucket_name)
    s3_client.put_bucket_versioning(
        Bucket=bucket_name, VersioningConfiguration={"Status": "Enabled"}
    )
    s3_client.put_bucket_replication(
        Bucket=bucket_name,
        ReplicationConfiguration={
            "Role": "myrole",
            "Rules": [
                {"Destination": {"Bucket": "secondbucket"}, "Status": "Enabled"},
                {
                    "ID": "secondrule",
                    "Priority": 2,
                    "Destination": {"Bucket": "thirdbucket"},
                    "Status": "Disabled",
                },
            ],
        },
    )

    config = s3_client.get_bucket_replication(Bucket=bucket_name)[
        "ReplicationConfiguration"
    ]
    assert config["Role"] == "myrole"
    rules = config["Rules"]
    assert len(rules) == 2

    first_rule = rules[0]
    assert "ID" in first_rule
    assert first_rule["Priority"] == 1
    assert first_rule["Status"] == "Enabled"
    assert first_rule["Destination"] == {"Bucket": "secondbucket"}

    second = rules[1]
    assert second["ID"] == "secondrule"
    assert second["Priority"] == 2
    assert second["Status"] == "Disabled"
    assert second["Destination"] == {"Bucket": "thirdbucket"}
